# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides reference systems for brains."""

from typing import List

from . import concept, space as _space, parcellation as _parcellation
from ..commons import MapType, logger, InstanceTable, Species


VERSION_BLACKLIST_WORDS = ["beta", "rc", "alpha"]


class Atlas(concept.AtlasConcept, configuration_folder="atlases"):
    """
    Main class for an atlas, providing access to feasible
    combinations of available parcellations and reference
    spaces, as well as common functionalities of those.
    """

    def __init__(self, identifier: str, name: str, species: Species, **kwargs):
        """Construct an empty atlas object with a name and identifier."""

        concept.AtlasConcept.__init__(
            self,
            identifier=identifier,
            name=name,
            species=species,
            **kwargs
        )
        self._parcellation_ids: List[str] = []
        self._space_ids: List[str] = []

    def _register_space(self, space_id: str):
        self._space_ids.append(space_id)

    def _register_parcellation(self, parcellation_id: str):
        self._parcellation_ids.append(parcellation_id)

    @property
    def spaces(self):
        """Access a registry of reference spaces supported by this atlas."""
        return InstanceTable[_space.Space](
            elements={s.key: s for s in _space.Space.registry() if s.id in self._space_ids},
            matchfunc=_space.Space.match,
        )

    @property
    def parcellations(self):
        """Access a registry of parcellations supported by this atlas."""
        return InstanceTable[_parcellation.Parcellation](
            elements={p.key: p for p in _parcellation.Parcellation.registry() if p.id in self._parcellation_ids},
            matchfunc=_parcellation.Parcellation.match,
        )

    def get_parcellation(self, parcellation=None) -> "_parcellation.Parcellation":
        """
        Returns a valid parcellation object defined by the atlas. If no
        specification is provided, the default is returned.

        Parameters
        ----------
        parcellation: str, Parcellation
            specification of a parcellation or a parcellation object

        Returns
        -------
            Parcellation
        """

        if parcellation is None:
            parcellation_obj = self.parcellations[self._parcellation_ids[0]]
            if len(self._parcellation_ids) > 1:
                logger.info(f"No parcellation specified, using default: '{parcellation_obj.name}'.")
            return parcellation_obj

        if isinstance(parcellation, _parcellation.Parcellation):
            assert parcellation in self.parcellations
            return parcellation

        return self.parcellations[parcellation]

    def get_space(self, space=None) -> "_space.Space":
        """
        Returns a valid reference space object defined by the atlas. If no
        specification is provided, the default is returned.

        Parameters
        ----------
        space: str, Space
            specification of a space or a space object

        Returns
        -------
            Space
        """
        if space is None:
            space_obj = self.spaces[self._space_ids[0]]
            if len(self._space_ids) > 1:
                logger.info(f"No space specified, using default '{space_obj.name}'.")
            return space_obj

        if isinstance(space, _space.Space):
            assert space in self.spaces
            return space

        return self.spaces[space]

    def get_map(
        self,
        space: _space.Space = None,
        parcellation: _parcellation.Parcellation = None,
        maptype: MapType = MapType.LABELLED,
    ):
        """
        Returns a parcellation map in the given space.

        Parameters
        ----------

        space: Space
            The requested reference space. If None, the default is used.
        parcellation: Parcellation
            The requested parcellation. If None, the default is used.
        maptype: MapType
            Type of the map (labelled or statistical/probabilistic)

        Returns
        -------
        ParcellationMap
        """
        parc_obj = self.get_parcellation(parcellation)
        space_obj = self.get_space(space)
        return parc_obj.get_map(space=space_obj, maptype=maptype)

    def get_region(self, region, parcellation=None):
        """
        Returns a valid Region object matching the given specification.

        Parameters
        ----------
        region : str or Region
            Key, approximate name, id or instance of a brain region
        parcellation : str or Parcellation
            Key, approximate name, id or instance of a brain parcellation.
            If None, the default is used.
        """
        return self.get_parcellation(parcellation).get_region(region)

    def get_template(self, space: _space.Space = None, variant: str = None):
        """
        Returns the reference template in the desired reference space.
        If no reference space is given, the default from `Atlas.space()` is used.

        Parameters
        ----------
        space: Space
            The desired reference space
        variant: str (optional)
            Some templates are provided in different variants, e.g.
            freesurfer is available as either white matter, pial or
            inflated surface for left and right hemispheres (6 variants).
            This field could be used to request a specific variant.
            Per default, the first found variant is returned.
        """
        return self.get_space(space).get_template(variant=variant)

    def get_voi(self, space: _space.Space, point1: tuple, point2: tuple):
        """Get a volume of interest spanned by two points in the given reference space.

        Parameters
        ----------
        space: Space, str
            The target reference space, or a string specification of the space
        point1: Tuple
            A 3D coordinate given in this reference space
        point2: Tuple
            Another 3D coordinate given in this reference space

        Returns
        -------
            BoundingBox
        """
        return self.get_template(space).get_boundingbox(point1, point2)

    def find_regions(
        self,
        regionspec: str,
        all_versions: bool = False,
        filter_children: bool = True,
        find_topmost: bool = False
    ):
        """
        Find regions with the given specification in all parcellations offered
        by the atlas.

        Parameters
        ----------
        regionspec: str, regex
            - a string with a possibly inexact name (matched both against the name and the identifier key)
            - a string in '/pattern/flags' format to use regex search (acceptable flags: aiLmsux, see at https://docs.python.org/3/library/re.html#flags)
            - a regex applied to region names
        all_versions : Bool, default: False
            If True, matched regions for all versions of a parcellation are returned.
        filter_children : bool, default: True
            If False, children of matched parents will be returned.
        find_topmost : bool, default: False
            If True (requires `filter_children=True`), will return parent
            structures if all children are matched, even though the parent
            itself might not match the specification.

        Returns
        -------
        list[Region]
            list of regions matching to the regionspec
        """
        result = []
        for p in self.parcellations:
            if p.is_newest_version or all_versions:
                result.extend(
                    p.find(
                        regionspec=regionspec,
                        filter_children=filter_children,
                        find_topmost=find_topmost
                    )
                )
        return result



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A specific mesh or 3D array."""

from typing import List, Dict, Union, Set, TYPE_CHECKING
from dataclasses import dataclass
from time import sleep
import json
from functools import lru_cache

import numpy as np
from nibabel import Nifti1Image
from skimage import feature as skimage_feature, filters

from . import providers as _providers
from ..commons import (
    resample_img_to_img,
    siibra_tqdm,
    affine_scaling,
    connected_components,
    logger,
    QUIET,
)
from ..exceptions import NoMapAvailableError, SpaceWarpingFailedError, EmptyPointCloudError
from ..retrieval import requests
from ..core import space as _space, structure
from ..core.concept import get_registry
from ..locations import point, pointcloud, boundingbox

if TYPE_CHECKING:
    from ..retrieval.datasets import EbrainsDataset as TypeDataset


@dataclass
class ComponentSpatialProperties:
    """
    Centroid and nonzero volume of an image.
    """
    centroid: point.Point
    volume: int

    @staticmethod
    def compute_from_image(
        img: Nifti1Image,
        space: Union[str, "_space.Space"],
        split_components: bool = True

    ) -> List["ComponentSpatialProperties"]:
        """
        Find the center of an image in its (non-zero) voxel space and and its
        volume.

        Parameters
        ----------
        img: Nifti1Image
        space: str, Space
        split_components: bool, default: True
            If True, finds the spatial properties for each connected component
            found by skimage.measure.label.
        """
        scale = affine_scaling(img.affine)
        if split_components:
            iter_components = lambda img: connected_components(
                np.asanyarray(img.dataobj),
                connectivity=None
            )
        else:
            iter_components = lambda img: [(0, np.asanyarray(img.dataobj))]

        spatial_props: List[ComponentSpatialProperties] = []
        for _, component in iter_components(img):
            nonzero: np.ndarray = np.c_[np.nonzero(component)]
            spatial_props.append(
                ComponentSpatialProperties(
                    centroid=point.Point(
                        np.dot(img.affine, np.r_[nonzero.mean(0), 1])[:3],
                        space=space
                    ),
                    volume=nonzero.shape[0] * scale,
                )
            )

        # sort by volume
        spatial_props.sort(key=lambda cmp: cmp.volume, reverse=True)

        return spatial_props


class Volume(structure.BrainStructure):
    """
    A volume is a specific mesh or 3D array,
    which can be accessible via multiple providers in different formats.
    """

    IMAGE_FORMATS = [
        "nii",
        "zip/nii",
        "neuroglancer/precomputed"
    ]

    MESH_FORMATS = [
        "neuroglancer/precompmesh",
        "neuroglancer/precompmesh/surface",
        "gii-mesh",
        "gii-label",
        "freesurfer-annot",
        "zip/freesurfer-annot",
    ]

    SUPPORTED_FORMATS = IMAGE_FORMATS + MESH_FORMATS

    _FORMAT_LOOKUP = {
        "image": IMAGE_FORMATS,
        "mesh": MESH_FORMATS,
        "surface": MESH_FORMATS,
        "nifti": ["nii", "zip/nii"],
        "nii": ["nii", "zip/nii"]
    }

    _FETCH_CACHE = {}  # we keep a cache of the most recently fetched volumes
    _FETCH_CACHE_MAX_ENTRIES = 3

    def __init__(
        self,
        space_spec: dict,
        providers: List[_providers.provider.VolumeProvider],
        name: str = "",
        variant: str = None,
        datasets: List['TypeDataset'] = [],
        bbox: "boundingbox.BoundingBox" = None
    ):
        self._name = name
        self._space_spec = space_spec
        self.variant = variant
        self._providers: Dict[str, _providers.provider.VolumeProvider] = {}
        self.datasets = datasets
        self._boundingbox = bbox
        for provider in providers:
            srctype = provider.srctype
            assert srctype not in self._providers
            self._providers[srctype] = provider
        if len(self._providers) == 0:
            logger.debug(f"No provider for volume {name}")

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other: 'Volume'):
        return (
            isinstance(other, Volume)
            and self.name == other.name
            and self.space == other.space
            and self.variant == other.variant
            and self._providers == other._providers
        )

    @property
    def name(self):
        """Allows derived classes to implement a lazy name specification."""
        return self._name

    @property
    def providers(self):
        def concat(url: Union[str, Dict[str, str]], concat: str):
            if isinstance(url, str):
                return url + concat
            return {key: url[key] + concat for key in url}
        return {
            srctype: concat(prov._url, f" {prov.label}" if hasattr(prov, "label") else "")
            for srctype, prov in self._providers.items()
        }

    @lru_cache(2)
    def get_boundingbox(self, clip: bool = True, background: float = 0.0, **fetch_kwargs) -> "boundingbox.BoundingBox":
        """
        Obtain the bounding box in physical coordinates of this volume.

        Parameters
        ----------
        clip : bool, default: True
            Whether to clip the background of the volume.
        background : float, default: 0.0
            The background value to clip.
            Note
            ----
            To use it, clip must be True.
        fetch_kwargs:
            key word arguments that are used for fetching volumes,
            such as voi or resolution_mm. Currently, only possible for
            Neuroglancer volumes except for `format`.

        Raises
        ------
        RuntimeError
            If the volume provider does not have a bounding box calculator.
        """
        if self._boundingbox is not None and len(fetch_kwargs) == 0:
            return self._boundingbox

        if not self.provides_image:
            raise NotImplementedError("Bounding box calculation of meshes is not implemented yet.")

        if clip:  # clipping requires fetching the image
            img = self.fetch(**fetch_kwargs)
            assert isinstance(img, Nifti1Image)
            return boundingbox.from_array(
                array=np.asanyarray(img.dataobj),
                background=background,
            ).transform(img.affine, space=self.space)

        # if clipping is not required, providers might have methods of creating
        # bounding boxes without fetching the image
        fmt = fetch_kwargs.get("format")
        if (fmt is not None) and (fmt not in self.formats):
            raise ValueError(
                f"Requested format {fmt} is not available as provider of "
                "this volume. See `volume.formats` for possible options."
            )
        providers = [self._providers[fmt]] if fmt else self._providers.values()
        for provider in providers:
            try:
                assert clip is False
                bbox = provider.get_boundingbox(
                    background=background, **fetch_kwargs
                )
                if bbox.space is None:  # provider do not know the space!
                    bbox._space_cached = self.space
                    bbox.minpoint._space_cached = self.space
                    bbox.maxpoint._space_cached = self.space
            except NotImplementedError:
                continue
            return bbox
        raise RuntimeError(f"No bounding box specified by any volume provider of {str(self)}")

    @property
    def formats(self) -> Set[str]:
        return {fmt for fmt in self._providers}

    @property
    def provides_mesh(self):
        return any(f in self.MESH_FORMATS for f in self.formats)

    @property
    def provides_image(self):
        return any(f in self.IMAGE_FORMATS for f in self.formats)

    @property
    def fragments(self) -> Dict[str, List[str]]:
        result = {}
        for srctype, p in self._providers.items():
            t = 'mesh' if srctype in self.MESH_FORMATS else 'image'
            for fragment_name in p.fragments:
                if t in result:
                    result[t].append(fragment_name)
                else:
                    result[t] = [fragment_name]
        return result

    @property
    def space(self):
        for key in ["@id", "name"]:
            if key in self._space_spec:
                return _space.Space.get_instance(self._space_spec[key])
        return _space.Space(None, "Unspecified space", species=_space.Species.UNSPECIFIED_SPECIES)

    @property
    def species(self):
        s = self.space
        return None if s is None else s.species

    def __str__(self):
        return (
            f"{self.__class__.__name__} {f'{self.name}' if self.name else ''}"
            f"{f' in space {self.space.name}' if self.space else ''}"
        )

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}(space_spec={self._space_spec}, "
            f"name='{self.name}', providers={self._providers})>"
        )

    def evaluate_points(
        self,
        points: Union['point.Point', 'pointcloud.PointCloud'],
        outside_value: Union[int, float] = 0,
        **fetch_kwargs
    ) -> np.ndarray:
        """
        Evaluate the image at the positions of the given points.

        Note
        ----
        Uses nearest neighbor interpolation. Other interpolation schemes are not
        yet implemented.

        Note
        ----
        If points are not on the same space as the map, they will be warped to
        the space of the volume.

        Parameters
        ----------
        points: PointCloud
        outside_value: int, float. Default: 0
        fetch_kwargs: dict
            Any additional arguments are passed to the `fetch()` call for
            retrieving the image data.

        Returns
        -------
        values: numpy.ndarray
            The values of the volume at the voxels points correspond to.

        Raises
        ------
        SpaceWarpingFailedError
            If warping of the points fails.
        """
        if not self.provides_image:
            raise NotImplementedError("Filtering of points by pure mesh volumes not yet implemented.")

        # make sure the points are in the same physical space as this volume
        as_pointcloud = pointcloud.from_points([points]) if isinstance(points, point.Point) else points
        warped = as_pointcloud.warp(self.space)
        assert warped is not None, SpaceWarpingFailedError

        # get the voxel array of this volume
        img = self.fetch(format='image', **fetch_kwargs)
        arr = np.asanyarray(img.dataobj)

        # transform the points to the voxel space of the volume for extracting values
        phys2vox = np.linalg.inv(img.affine)
        voxels = warped.transform(phys2vox, space=None)
        XYZ = voxels.coordinates.astype('int')

        # temporarily set all outside voxels to (0,0,0) so that the index access doesn't fail
        inside = np.all((XYZ < arr.shape) & (XYZ > 0), axis=1)
        XYZ[~inside, :] = 0

        # read out the values
        X, Y, Z = XYZ.T
        values = arr[X, Y, Z]

        # fix the outside voxel values, which might have an inconsistent value now
        values[~inside] = outside_value

        return values

    def _points_inside(
        self,
        points: Union['point.Point', 'pointcloud.PointCloud'],
        keep_labels: bool = True,
        outside_value: Union[int, float] = 0,
        **fetch_kwargs
    ) -> 'pointcloud.PointCloud':
        """
        Reduce a pointcloud to the points which fall inside nonzero pixels of
        this map.


        Parameters
        ----------
        points: PointCloud
        keep_labels: bool
            If False, the returned PointCloud will be labeled with their indices
            in the original PointCloud.
        fetch_kwargs: dict
            Any additional arguments are passed to the `fetch()` call for
            retrieving the image data.

        Returns
        -------
        PointCloud
            A new PointCloud containing only the points inside the volume.
            Labels reflect the indices of the original points if `keep_labels`
            is False.
        """
        ptset = pointcloud.from_points([points]) if isinstance(points, point.Point) else points
        values = self.evaluate_points(ptset, outside_value=outside_value, **fetch_kwargs)
        inside = list(np.where(values != outside_value)[0])
        return pointcloud.from_points(
            [ptset[i] for i in inside],
            newlabels=None if keep_labels else inside
        )

    def intersection(self, other: structure.BrainStructure, **fetch_kwargs) -> structure.BrainStructure:
        """
        Compute the intersection of a location with this volume. This will
        fetch actual image data. Any additional arguments are passed to fetch.
        """
        if isinstance(other, (pointcloud.PointCloud, point.Point)):
            try:
                points_inside = self._points_inside(other, keep_labels=False, **fetch_kwargs)
            except EmptyPointCloudError:
                return None  # BrainStructure.intersects checks for not None
            if isinstance(other, point.Point):  # preserve the type
                return points_inside[0]
            return points_inside
        elif isinstance(other, boundingbox.BoundingBox):
            return self.get_boundingbox(clip=True, background=0.0, **fetch_kwargs).intersection(other)
        elif isinstance(other, Volume):
            if self.space != other.space:
                raise NotImplementedError("Cannot intersect volumes from different spaces. Try comparing their bounding boxes.")
            format = fetch_kwargs.pop('format', 'image')
            v1 = self.fetch(format=format, **fetch_kwargs)
            v2 = other.fetch(format=format, **fetch_kwargs)
            arr1 = np.asanyarray(v1.dataobj)
            arr2 = np.asanyarray(resample_img_to_img(v2, v1).dataobj)
            pointwise_min = np.minimum(arr1, arr2)
            if np.any(pointwise_min):
                return from_array(
                    data=pointwise_min,
                    affine=v1.affine,
                    space=self.space,
                    name=f"Intersection between {self} and {other} computed as their pointwise minimum"
                )
            else:
                return None
        else:  # other BrainStructures should have intersection with locations implemented.
            try:
                return other.intersection(self)
            except NoMapAvailableError:
                return None

    def fetch(self, format: str = None, **kwargs):
        """
        Fetch a volumetric or surface representation from one of the providers.

        Parameters
        ----------
        format: str, default=None
            Requested format. If `None`, the first supported format matching in
            `self.formats` is tried, starting with volumetric formats.
            It can be explicitly specified as:
                - 'surface' or 'mesh' to fetch a surface format
                - 'volumetric' or 'voxel' to fetch a volumetric format
                - supported format types, see SUPPORTED_FORMATS. This includes
                'nii', 'zip/nii', 'neuroglancer/precomputed', 'gii-mesh',
                'neuroglancer/precompmesh', 'gii-label'
        **kwargs
            - resolution_mm: resolution in millimeters as float or a tuple of floats
            - format: the format of the volume, like "mesh" or "nii"
            - voi: a BoundingBox of interest

        Returns
        -------
        An image (Nifti1Image) or mesh (Dict['verts': ndarray, 'faces': ndarray, 'labels': ndarray])
        """
        kwargs_serialized = json.dumps({k: hash(v) for k, v in kwargs.items()}, sort_keys=True)

        if "resolution_mm" in kwargs and format is None:
            if 'neuroglancer/precomputed' not in self.formats:
                raise ValueError("'resolution_mm' is only available for volumes with 'neuroglancer/precomputed' formats.")
            format = 'neuroglancer/precomputed'

        if format is None:
            # preserve fetch order in SUPPORTED_FORMATS
            possible_formats = [f for f in self.SUPPORTED_FORMATS if f in self.formats]
        elif format in self._FORMAT_LOOKUP:  # allow use of aliases
            possible_formats = [f for f in self._FORMAT_LOOKUP[format] if f in self.formats]
        elif format in self.SUPPORTED_FORMATS:
            possible_formats = [format] if format in self.formats else []
        else:
            possible_formats = []
        if len(possible_formats) == 0:
            raise ValueError(
                f"Invalid format requested: {format}. Possible values for this "
                f"volume are: {self.formats}"
            )

        # ensure the voi is inside the template
        voi = kwargs.get("voi", None)
        if voi is not None and voi.space is not None:
            assert isinstance(voi, boundingbox.BoundingBox)
            tmplt_bbox = voi.space.get_template().get_boundingbox(clip=False)
            intersection_bbox = voi.intersection(tmplt_bbox)
            if intersection_bbox is None:
                raise RuntimeError(f"voi provided ({voi}) lies out side the voxel space of the {voi.space.name} template.")
            if intersection_bbox != voi:
                logger.info(
                    f"Since provided voi lies outside the template ({voi.space}) it is clipped as: {intersection_bbox}"
                )
                kwargs["voi"] = intersection_bbox

        result = None
        # try each possible format
        for fmt in possible_formats:
            fetch_hash = hash((hash(self), hash(fmt), hash(kwargs_serialized)))
            # cached
            if fetch_hash in self._FETCH_CACHE:
                break
            # Repeat in case of too many requests only
            fwd_args = {k: v for k, v in kwargs.items() if k != "format"}
            for try_count in range(6):
                try:
                    if fmt in ["gii-label", "freesurfer-annot", "zip/freesurfer-annot"]:
                        tpl = self.space.get_template(variant=kwargs.get('variant'))
                        mesh = tpl.fetch(**kwargs)
                        labels = self._providers[fmt].fetch(**fwd_args)
                        result = dict(**mesh, **labels)
                    else:
                        result = self._providers[fmt].fetch(**fwd_args)
                except requests.SiibraHttpRequestError as e:
                    if e.status_code == 429:  # too many requests
                        sleep(0.1)
                        logger.error(f"Cannot access {self._providers[fmt]}", exc_info=(try_count == 5))
                        continue
                    else:
                        break
                except Exception as e:
                    logger.info(e, exc_info=1)
                    break
                else:
                    break
            # update the cache if fetch is successful
            if result is not None:
                self._FETCH_CACHE[fetch_hash] = result
                while len(self._FETCH_CACHE) >= self._FETCH_CACHE_MAX_ENTRIES:
                    # remove oldest entry
                    self._FETCH_CACHE.pop(next(iter(self._FETCH_CACHE)))
                break
        else:
            # unsuccessful: do not poison the cache if none fetched
            logger.error(f"Could not fetch any formats from {possible_formats}.")
            return None

        return self._FETCH_CACHE[fetch_hash]

    def fetch_connected_components(self, **fetch_kwargs):
        """
        Provide an generator over masks of connected components in the volume
        """
        img = self.fetch(**fetch_kwargs)
        assert isinstance(img, Nifti1Image), NotImplementedError(
            f"Connected components for type {type(img)} is not yet implemented."
        )
        for label, component in connected_components(np.asanyarray(img.dataobj)):
            yield (
                label,
                Nifti1Image(component, img.affine)
            )

    def compute_spatial_props(self, split_components: bool = True, **fetch_kwargs) -> List[ComponentSpatialProperties]:
        """
        Find the center of this volume in its (non-zero) voxel space and and its
        volume.

        Parameters
        ----------
        split_components: bool, default: True
            If True, finds the spatial properties for each connected component
            found by skimage.measure.label.
        """
        assert self.provides_image, NotImplementedError("Spatial properties can currently on be calculated for images.")
        img = self.fetch(format=fetch_kwargs.pop("format", "image"), **fetch_kwargs)
        return ComponentSpatialProperties.compute_from_image(
            img=img,
            space=self.space,
            split_components=split_components
        )

    def draw_samples(self, N: int, sample_size: int = 100, e: float = 1, sigma_mm=None, invert=False, **kwargs):
        """
        Draw samples from the volume, by interpreting its values as an
        unnormalized empirical probability distributions.
        Any keyword arguments are passed over to fetch()
        """
        if not self.provides_image:
            raise NotImplementedError(
                "Drawing samples is so far only implemented for image-type volumes, "
                f"not {self.__class__.__name__}."
            )
        img = self.fetch(**kwargs)
        array = np.asanyarray(img.dataobj)
        samples = []
        P = (array - array.min()) / (array.max() - array.min())
        if invert:
            P = 1 - P
        P = P**e
        while True:
            pts = (np.random.rand(sample_size, 3) * max(P.shape))
            inside = np.all(pts < P.shape, axis=1)
            Y, X, Z = np.split(pts[inside, :].astype('int'), 3, axis=1)
            T = np.random.rand(1)
            choice = np.where(P[Y, X, Z] >= T)[0]
            samples.extend(list(pts[inside, :][choice, :]))
            if len(samples) > N:
                break
        voxels = pointcloud.PointCloud(
            np.random.permutation(samples)[:N, :],
            space=None
        )
        result = voxels.transform(img.affine, space='mni152')
        result.sigma_mm = [sigma_mm for _ in result]
        return result

    def find_peaks(self, mindist=5, sigma_mm=0, **kwargs):
        """
        Find local peaks in the volume.
        Additional keyword arguments are passed over to fetch()
        """
        if not self.provides_image:
            raise NotImplementedError(
                "Finding peaks is so far only implemented for image-type volumes, "
                f"not {self.__class__.__name__}."
            )
        img = self.fetch(**kwargs)
        array = np.asanyarray(img.dataobj)
        voxels = skimage_feature.peak_local_max(array, min_distance=mindist)
        points = pointcloud.PointCloud(voxels, space=None, labels=list(range(len(voxels)))).transform(img.affine, space=self.space)
        points.sigma_mm = [sigma_mm for _ in points]
        return points


class FilteredVolume(Volume):

    def __init__(
        self,
        parent_volume: Volume,
        label: int = None,
        fragment: str = None,
        threshold: float = None,
    ):
        """
        A prescribed Volume to fetch specified label and fragment.
        If threshold is defined, a mask of the values above the threshold.

        Parameters
        ----------
        parent_volume : Volume
        label : int, default: None
            Get the mask of value equal to label.
        fragment : str, default None
            If a volume is fragmented, get a specified one.
        threshold : float, default None
            Provide a float value to threshold the image.
        """
        name = parent_volume.name
        if label:
            name += f" - label: {label}"
        if fragment:
            name += f" - fragment: {fragment}"
        if threshold:
            name += f" - threshold: {threshold}"
        Volume.__init__(
            self,
            space_spec=parent_volume._space_spec,
            providers=list(parent_volume._providers.values()),
            name=name
        )
        self.fragment = fragment
        self.label = label
        self.threshold = threshold

    def fetch(
        self,
        format: str = None,
        **kwargs
    ):
        if "fragment" in kwargs:
            assert kwargs.get("fragment") == self.fragment, f"This is a filtered volume that can only fetch fragment '{self.fragment}'."
        else:
            kwargs["fragment"] = self.fragment
        if "label" in kwargs:
            assert kwargs.get("label") == self.label, f"This is a filtered volume that can only fetch label '{self.label}' only."
        else:
            kwargs["label"] = self.label

        result = super().fetch(format=format, **kwargs)

        if self.threshold is not None:
            assert self.label is None
            if not isinstance(result, Nifti1Image):
                raise NotImplementedError("Cannot threshold meshes.")
            imgdata = np.asanyarray(result.dataobj)
            return Nifti1Image(
                dataobj=(imgdata > self.threshold).astype("uint8"),
                affine=result.affine,
                dtype="uint8"
            )

        return result

    def get_boundingbox(
        self,
        clip: bool = True,
        background: float = 0.0,
        **fetch_kwargs
    ) -> "boundingbox.BoundingBox":
        # NOTE: since some providers enable different simpllified ways to create a
        # bounding box without fetching the image, the correct kwargs must be
        # forwarded since FilteredVolumes enforce their specs to be fetched.
        return super().get_boundingbox(
            clip=clip,
            background=background,
            **fetch_kwargs
        )


class ReducedVolume(Volume):
    def __init__(
        self,
        source_volumes: List[Volume],
        new_labels: List[int] = None,
    ):
        """
        A prescribed Volume to fetch a list of source volumes in the same space,
        with optionally new labels, and resampling them to the template of this
        space to be merged into one image.

        Parameters
        ----------
        source_volumes: List[Volume]
        new_labels: List[int], default: None
        """
        assert len({v.space for v in source_volumes}) == 1, "Cannot merge volumes from different spaces."
        if new_labels:
            assert len(source_volumes) == len(new_labels), "Need to supply as many labels as volumes."
        Volume.__init__(
            self,
            space_spec=source_volumes[0]._space_spec,
            providers=[],
            name=f"Volumes resampled to {source_volumes[0].space} template and merged: {','.join([v.name for v in source_volumes])}"
        )
        self.source_volumes = source_volumes
        self.new_labels = new_labels

    @lru_cache(2)
    def fetch(self, format: str = None, **kwargs):
        # determine dtype
        if self.new_labels is not None:
            if set(self.new_labels) != {1}:
                logger.info(f"Relabling regions with labels: {self.new_labels}")
            dtype = 'int32'
        elif {type(v) for v in self.source_volumes} == {FilteredVolume}:
            dtype = 'uint8'
        else:
            dtype = self.source_volumes[0].fetch().dataobj.dtype

        # determine base image
        template_img = self.space.get_template().fetch(format=format, **kwargs)
        if format in [None, "image"] and "neuroglancer/precomputed" in self.formats:
            format = "neuroglancer/precomputed"
            template_res = min(template_img.header.get_zooms())
            kwargs["resolution_mm"] = kwargs.get("resolution_mm", template_res)

        merged_array = np.zeros(template_img.shape, dtype=dtype)
        for i, vol in siibra_tqdm(
            enumerate(self.source_volumes),
            unit=" volume",
            desc=f"Resampling volumes to {self.space.name} and merging",
            total=len(self.source_volumes),
            disable=len(self.source_volumes) < 3,
            leave=False,
        ):
            img = vol.fetch(format=format, **kwargs)
            if img is None:
                continue
            resampled_arr = np.asanyarray(
                resample_img_to_img(img, template_img).dataobj
            )
            nonzero_voxels = resampled_arr > 0
            if self.new_labels:
                merged_array[nonzero_voxels] = self.new_labels[i]
            else:
                merged_array[nonzero_voxels] = resampled_arr[nonzero_voxels]

        result = Nifti1Image(merged_array, affine=template_img.affine, dtype=dtype)
        if 'neuroglancer/precomputed' in self.formats:
            result.set_qform(result.affine)
        return result

    def get_boundingbox(
        self,
        clip: bool = True,
        background: float = 0.0,
        **fetch_kwargs
    ) -> boundingbox.BoundingBox:
        if clip:
            with QUIET:
                return super().get_boundingbox(clip=clip, background=background, **fetch_kwargs)
        else:
            return boundingbox.BoundingBox.union(
                *[
                    v.get_boundingbox(clip=clip, background=background, **fetch_kwargs)
                    for v in self.source_volumes
                ]
            )

    @property
    def formats(self) -> Set[str]:
        return {fmt for v in self.source_volumes for fmt in v._providers}


class Subvolume(Volume):
    """
    Wrapper class for exposing a z level of a 4D volume to be used like a 3D volume.
    """

    def __init__(self, parent_volume: Volume, z: int):
        Volume.__init__(
            self,
            space_spec=parent_volume._space_spec,
            providers=[
                _providers.provider.SubvolumeProvider(p, z=z)
                for p in parent_volume._providers.values()
            ],
            name=parent_volume.name + f" - z: {z}"
        )


def from_file(filename: str, space: str, name: str) -> Volume:
    """Builds a nifti volume from a filename."""
    spaceobj = get_registry("Space").get(space)
    return Volume(
        space_spec={"@id": spaceobj.id},
        providers=[_providers.NiftiProvider(filename)],
        name=filename if name is None else name,
    )


def from_nifti(nifti: Nifti1Image, space: str, name: str) -> Volume:
    """Builds a nifti volume from a Nifti image."""
    spaceobj = get_registry("Space").get(space)
    return Volume(
        space_spec={"@id": spaceobj.id},
        providers=[_providers.NiftiProvider((np.asanyarray(nifti.dataobj), nifti.affine))],
        name=name
    )


def from_array(
    data: np.ndarray,
    affine: np.ndarray,
    space: Union[str, Dict[str, str]],
    name: str
) -> Volume:
    """Builds a siibra volume from an array and an affine matrix."""
    if len(name) == 0:
        raise ValueError("Please provide a non-empty string for `name`")
    spacespec = next(iter(space.values())) if isinstance(space, dict) else space
    spaceobj = get_registry("Space").get(spacespec)
    return Volume(
        space_spec={"@id": spaceobj.id},
        providers=[_providers.NiftiProvider((data, affine))],
        name=name,
    )


def from_pointcloud(
    points: pointcloud.PointCloud,
    label: int = None,
    target: Volume = None,
    normalize=True,
    **kwargs
) -> Volume:
    """
    Get the kernel density estimate as a volume from the points using their
    average uncertainty on target volume.

    Parameters
    ----------
    points: pointcloud.PointCloud
    label: int, default: None
        If None, finds the KDE for all points. Otherwise, selects the points
        labelled with this integer value.
    target: Volume, default: None
        If None, the template of the space points are defined on will be used.
    normalize: bool, default: True

    Raises
    ------
    RuntimeError
        If no points with labels found
    """
    if target is None:
        target = points.space.get_template()
    targetimg = target.fetch(**kwargs)
    voxels = points.transform(np.linalg.inv(targetimg.affine), space=None)

    if (label is None) or (points.labels is None):
        selection = [True for _ in points]
    else:
        assert label in points.labels, f"No points with the label {label} in the set: {set(points.labels)}"
        selection = points.labels == label

    voxelcount_img = np.zeros_like(targetimg.get_fdata())
    unique_coords, counts = np.unique(
        np.array(voxels.as_list(), dtype='int')[selection, :],
        axis=0,
        return_counts=True
    )
    voxelcount_img[tuple(unique_coords.T)] = counts

    # TODO: consider how to handle pointclouds with varied sigma_mm
    sigmas = np.array(points.sigma_mm)[selection]
    bandwidth = np.mean(sigmas)
    if len(np.unique(sigmas)) > 1:
        logger.warning(f"KDE of pointcloud uses average bandwidth {bandwidth} instead of the points' individual sigmas.")

    filtered_arr = filters.gaussian(voxelcount_img, bandwidth)
    if normalize:
        filtered_arr /= filtered_arr.sum()

    return from_array(
        data=filtered_arr,
        affine=targetimg.affine,
        space=target.space,
        name=f'KDE map of {points}{f"labelled {label}" if label else ""}'
    )


def merge(volumes: List[Volume], labels: List[int] = [], **fetch_kwargs) -> Volume:
    """
    Merge a list of nifti volumes in the same space into a single volume.

    Note
    ----
    In case of voxel conflicts, the volumes will be override the previous values
    in the given order.

    Parameters
    ----------
    volumes : List[Volume]
    labels : List[int], optional
        Supply new labels to replace existing values per volume.

    Returns
    -------
    Volume
    """
    if len(volumes) == 1:
        logger.debug("Only one volume supplied returning as is (kwargs are ignored).")
        return volumes[0]

    assert len(volumes) > 1, "Need to supply at least two volumes to merge."
    if labels:
        assert len(volumes) == len(labels), "Need to supply as many labels as volumes."

    space = volumes[0].space
    assert all(v.space == space for v in volumes), "Cannot merge volumes from different spaces."

    if len(labels) > 0:
        dtype = 'int32'
    elif FilteredVolume in {type(v) for v in volumes}:
        dtype = 'uint8'
    else:
        dtype = volumes[0].fetch().dataobj.dtype
    template_img = space.get_template().fetch(**fetch_kwargs)
    merged_array = np.zeros(template_img.shape, dtype=dtype)

    for i, vol in siibra_tqdm(
        enumerate(volumes),
        unit=" volume",
        desc=f"Resampling volumes to {space.name} and merging",
        total=len(volumes),
        disable=len(volumes) < 3
    ):
        img = vol.fetch(**fetch_kwargs)
        resampled_arr = np.asanyarray(
            resample_img_to_img(img, template_img).dataobj
        )
        nonzero_voxels = resampled_arr > 0
        if labels:
            merged_array[nonzero_voxels] = labels[i]
        else:
            merged_array[nonzero_voxels] = resampled_arr[nonzero_voxels]

    return from_array(
        data=merged_array,
        affine=template_img.affine,
        space=space,
        name=f"Resampled and merged volumes: {','.join([v.name for v in volumes])}"
    )



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Hierarchal brain regions and metadata."""

from functools import lru_cache
import re
from typing import Union, List, TYPE_CHECKING
try:
    from typing import Literal
except ImportError:
    # support python 3.7
    from typing_extensions import Literal

from . import region
from ..commons import logger, MapType, Species
from ..volumes import parcellationmap
from ..exceptions import MapNotFound


if TYPE_CHECKING:
    from .space import Space

# NOTE : such code could be used to automatically resolve
# multiple matching parcellations for a short spec to the newset version:
#               try:
#                    collections = {m.version.collection for m in matches}
#                    if len(collections)==1:
#                        return sorted(matches,key=lambda m:m.version,reverse=True)[0]
#                except Exception as e:
#                    pass


class ParcellationVersion:
    def __init__(
        self, name, parcellation, collection=None, prev_id=None, next_id=None, deprecated=False
    ):
        self.name = name
        self.collection = collection
        self.parcellation = parcellation
        self.next_id = next_id
        self.prev_id = prev_id
        self.deprecated = deprecated

    def __eq__(self, other):
        return all([self.name == other.name, self.collection == other.collection])

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __iter__(self):
        yield "name", self.name
        yield "prev", self.prev_id
        yield "next", self.next_id
        yield "deprecated", self.deprecated

    def __lt__(self, other: 'ParcellationVersion'):
        """
        < operator, useful for sorting by version
        FIXME: this is only by name, not recursing into parcellations, to avoid importing the registry here.
        """
        return self.name < other.name


class Parcellation(region.Region, configuration_folder="parcellations"):

    def __init__(
        self,
        identifier: str,
        name: str,
        species: Union[Species, str],
        regions: Union[List[region.Region], region.Region] = (),
        shortname: str = "",
        description: str = "",
        version: ParcellationVersion = None,
        modality: str = None,
        publications: list = [],
        datasets: list = [],
        prerelease: bool = False,
    ):
        """
        Constructs a new parcellation object.

        Parameters
        ----------
        identifier : str
            Unique identifier of the parcellation
        name : str
            Human-readable name of the parcellation
        species: str or Species
            Specification of the species
        regions: list or Region
        shortname: str
            Short form of human-readable name (optional)
        description: str
            Textual description of the parcellation
        version : str or None
            Version specification, optional
        modality  :  str or None
            Specification of the modality used for creating the parcellation
        publications: list
            List of associated publications, each a dictionary with "doi"
            and/or "citation" fields
        datasets : list
            datasets associated with this region
        """
        region.Region.__init__(
            self,
            name=name,
            children=regions,
            parent=None,
            shortname=shortname,
            description=description,
            publications=publications,
            datasets=datasets,
            modality=modality,
            prerelease=prerelease,
        )
        self._species_cached = Species.decode(species)
        self._id = identifier
        self.version = version

    @property
    def id(self):
        return self._id

    def matches(self, spec):
        if spec not in self._CACHED_MATCHES:
            if isinstance(spec, str):
                if all(
                    w in self.shortname.lower()
                    for w in re.split(r'\s+', spec.lower())
                ):
                    self._CACHED_MATCHES[spec] = True
        return super().matches(spec)

    def get_map(
        self,
        space: Union[str, "Space"],
        maptype: Union[Literal['labelled', 'statistical'], MapType] = MapType.LABELLED,
        spec: str = ""
    ):
        """
        Get the maps for the parcellation in the requested template space.

        This might in general include multiple 3D volumes. For example,
        the Julich-Brain atlas provides two separate maps, one per hemisphere.
        Per default, multiple maps are concatenated into a 4D array, but you
        can choose to retrieve a dict of 3D volumes instead using
        `return_dict=True`.

        Parameters
        ----------
        space: Space or str
            reference space specification such as name, id, or a `Space` instance.
        maptype: MapType or str
            Type of map requested (e.g., statistical or labelled).
            Use MapType.STATISTICAL to request probability maps.
            Defaults to MapType.LABELLED.
        spec: str, optional
            In case of multiple matching maps for the given parcellation, space
            and type, use this field to specify keywords matching the desired
            parcellation map name. Otherwise, siibra will default to the first
            in the list of matches (and inform with a log message)
        Returns
        -------
        parcellationmap.Map or SparseMap
            A ParcellationMap representing the volumetric map or
            a SparseMap representing the list of statistical maps.
        """
        if isinstance(maptype, str):
            maptype = MapType[maptype.upper()]
        assert isinstance(maptype, MapType), "Possible values of `maptype` are `MapType`s, 'labelled', 'statistical'."

        candidates = [
            m for m in parcellationmap.Map.registry()
            if m.space.matches(space)
            and m.maptype == maptype
            and m.parcellation.matches(self)
        ]
        if len(candidates) == 0:
            raise MapNotFound(f"No '{maptype}' map in '{space}' available for {str(self)}")
        if len(candidates) > 1:
            spec_candidates = [
                c for c in candidates if all(w.lower() in c.id.lower() for w in spec.split())
            ]
            if len(spec_candidates) == 0:
                raise MapNotFound(f"'{spec}' does not match any options from {[c.name for c in candidates]}.")
            if len(spec_candidates) > 1:
                logger.warning(
                    f"Multiple maps are available in this specification of space, parcellation, and map type.\n"
                    f"Choosing the first map from {[c.name for c in spec_candidates]}."
                )
            return spec_candidates[0]
        return candidates[0]

    @property
    def is_newest_version(self):
        return (self.version is None) or (self.version.next_id is None)

    def _split_group_spec(self, spec: str):
        """
        Split a group region specification as produced in older siibra versions
        into the subregion specs.
        This is used when decoding datasets that still include region name
        specifications from old siibra versions, such as some connectivity matrices.
        """
        spec = re.sub(r'Group: *', '', spec)
        for substr in re.findall(r'\(.*?\)', spec):
            # temporarily replace commas inside brackets with a placeholder
            # because these are not region spec delimiters
            spec = spec.replace(substr, re.sub(r', *', '##', substr))
        # process the comma separated substrings
        candidates = list({
            self.get_region(re.sub(r'##', ', ', s))
            for s in re.split(r', *', spec)
        })
        if len(candidates) > 0:
            return candidates
        else:
            return [spec]

    def get_region(
        self,
        regionspec: Union[str, region.Region],
        find_topmost: bool = False,
        allow_tuple: bool = False
    ):
        """
        Given a unique specification, return the corresponding region.

        The spec could be a (possibly incomplete) name, or a region object.
        This method is meant to definitely determine a valid region. Therefore,
        if no match is found, it raises a ValueError. If multiple matches are
        found, the method tries to return only the common parent node. If there
        is no common parent, an exception is raised, except when
        allow_tuple=True - then a tuple of matched regions is returned.

        Parameters
        ----------
        regionspec: str, Region
            - a string with a possibly inexact name (matched both against the name and the identifier key)
            - a Region object
        find_topmost: bool, default: False
            If True, will automatically return the parent of a decoded region
            the decoded region is its only child.
        allow_tuple: bool, default: False
            If multiple candidates without a common parent are found,
            return a tuple of matches instead of raising an exception.

        Returns
        -------
        Region
            A region object defined in the parcellation.

            Note
            ----
            If the spec exactly matched with more than one region, the first
            will be returned.

        Raises
        ------
        RuntimeError
            If the spec matches multiple regions
        ValueError
            If the spec cannot be matched against any region.
        """
        assert isinstance(regionspec, (str, region.Region)), f"get_region takes str or Region but you provided {type(regionspec)}"
        if isinstance(regionspec, region.Region) and (regionspec.parcellation == self):
            return regionspec

        # if there exist an exact match of region spec to region name, return
        if isinstance(regionspec, str):
            exact_match = [
                region
                for region in self
                if region.name == regionspec or region.key == regionspec
            ]
            if len(exact_match) == 1:
                return exact_match[0]
            if len(exact_match) > 1:
                logger.info(f"Found multiple region with exact match to {regionspec}. Returning the first one.")
                return exact_match[0]

        if regionspec.startswith("Group"):  # backwards compatibility with old "Group: <region a>, <region b>" specs
            candidates = {
                r
                for spec in self._split_group_spec(regionspec)
                for r in self.find(spec, filter_children=True, find_topmost=find_topmost)
            }
        else:
            candidates = self.find(regionspec, filter_children=True, find_topmost=find_topmost)

        if len(candidates) > 1 and isinstance(regionspec, str):
            # if we have an exact match of words in one region, discard other candidates.
            querywords = {w.replace(',', '').lower() for w in regionspec.split()}
            full_matches = []
            for c in candidates:
                targetwords = {w.lower() for w in c.name.split()}
                if len(querywords & targetwords) == len(targetwords):
                    full_matches.append(c)
            if len(full_matches) == 1:
                candidates = full_matches

        if not candidates:
            raise ValueError(f"'{regionspec}' could not be decoded under '{self.name}'")
        elif len(candidates) == 1:
            return candidates[0]
        else:
            if allow_tuple:
                return tuple(candidates)
            raise RuntimeError(
                f"Spec {regionspec!r} resulted in multiple matches: {', '.join(r.name for r in candidates)}."
            )

    def __getitem__(self, regionspec: Union[str, int]):
        """
        Retrieve a region object from the parcellation by labelindex or partial name.
        """
        return self.get_region(regionspec)

    def __lt__(self, other):
        """
        We sort parcellations by their version
        """
        if (self.version is None) or (other.version is None):
            logger.warning(
                f"Sorting non-versioned instances of {self.__class__.__name__} "
                f"by name: {self.name}, {other.name}"
            )
            return self.name < other.name
        return self.version.__lt__(other.version)


@lru_cache(maxsize=128)
def find_regions(
    regionspec: str,
    filter_children=True,
    find_topmost=False
):
    """
    Find regions matching the given region specification across all parcellation
    instances in the registry.

    Parameters
    ----------
    regionspec: str, regex, Region
        - a string with a possibly inexact name (matched both against the name and the identifier key)
        - a string in '/pattern/flags' format to use regex search (acceptable flags: aiLmsux) (see https://docs.python.org/3/library/re.html#flags)
        - a regex applied to region names
        - a Region object
    filter_children : bool, default: True
        If True, children of matched parents will not be returned
    find_topmost : bool, default: False
        If True (requires `filter_children=True`), will return parent
        structures if all children are matched, even though the parent
        itself might not match the specification.

    Returns
    -------
    list[Region]
        list of regions matching to the regionspec
    """
    result = []
    for p in Parcellation.registry():
        result.extend(
            p.find(
                regionspec=regionspec,
                filter_children=filter_children,
                find_topmost=find_topmost
            )
        )
    return result



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import wrap
from typing import List

import numpy as np
import pandas as pd

from . import tabular
from .. import anchor as _anchor
from ...commons import logger
from ...vocabularies import RECEPTOR_SYMBOLS


class ReceptorDensityFingerprint(
    tabular.Tabular,
    configuration_folder="features/tabular/fingerprints/receptor",
    category='molecular'
):

    DESCRIPTION = (
        "Fingerprint of densities (in fmol/mg protein) of receptors for classical neurotransmitters "
        "obtained by means of quantitative in vitro autoradiography. The fingerprint provides average "
        "density measurements for different receptors measured in tissue samples from different subjects "
        "together with the corresponding standard deviations. "
    )

    def __init__(
        self,
        tsvfile: str,
        anchor: _anchor.AnatomicalAnchor,
        datasets: list = [],
        id: str = None,
        prerelease: bool = False,
    ):
        """ Generate a receptor fingerprint from a URL to a .tsv file
        formatted according to the structure used by Palomero-Gallagher et al.
        """
        tabular.Tabular.__init__(
            self,
            description=self.DESCRIPTION,
            modality="Neurotransmitter receptor density",
            anchor=anchor,
            file=tsvfile,
            data=None,  # lazy loading below
            datasets=datasets,
            id=id,
            prerelease=prerelease,
        )

    @property
    def unit(self) -> str:
        return self._loader.data.iloc[:, -1][0]

    @property
    def receptors(self) -> List[str]:
        return list(self.data.index)

    @property
    def neurotransmitters(self) -> List[str]:
        # TODO quite a lot of receptor features have undecipherable symbols, mainly double quoted receptor
        # Likely ill-formed tsv's
        return [
            "{} ({})".format(
                RECEPTOR_SYMBOLS[t]['neurotransmitter']['label'],
                RECEPTOR_SYMBOLS[t]['neurotransmitter']['name'],
            ) if t in RECEPTOR_SYMBOLS else
            f"{t} (undeciphered)"
            for t in self.receptors
        ]

    @property
    def data(self):
        if self._data_cached is None:
            label_col, mean_col, std_col = list(self._loader.data.columns)[:3]
            self._data_cached = pd.DataFrame(
                np.array([
                    self._loader.data[mean_col],
                    self._loader.data[std_col]
                ]).T,
                index=self._loader.data[label_col],
                columns=['mean', 'std']
            )
            self._data_cached.index.name = 'receptor'
        return self._data_cached.copy()

    @classmethod
    def parse_tsv_data(cls, data: dict):
        units = {list(v.values())[3] for v in data.values()}
        labels = list(data.keys())
        assert len(units) == 1
        try:
            mean = [data[_]["density (mean)"] for _ in labels]
            std = [data[_]["density (sd)"] for _ in labels]
        except KeyError as e:
            print(str(e))
            logger.error("Could not parse fingerprint from this dictionary")
        return {
            'unit': next(iter(units)),
            'labels': labels,
            'means': [float(m) if m.isnumeric() else 0 for m in mean],
            'stds': [float(s) if s.isnumeric() else 0 for s in std],
        }

    def polar_plot(self, *args, backend='matplotlib', **kwargs):
        """
        Create a polar plot of the fingerprint.
        backend: str
            "matplotlib" or "plotly"
        """
        if backend == "matplotlib":
            try:
                import matplotlib.pyplot as plt
            except ImportError as e:
                logger.error(
                    "matplotlib not available. Please install matplotlib or use or another backend such as plotly."
                )
                raise e
            from collections import deque

            # default args
            wrapwidth = 40
            y = kwargs.pop("y") if "y" in kwargs else self.data.columns[0]
            yerr = kwargs.pop("yerr") if "yerr" in kwargs else None
            if yerr is None:
                yerr = 'std' if 'std' in self.data.columns else None
            ax = kwargs.pop("ax") if "ax" in kwargs else plt.subplot(111, projection="polar")

            datafield = y or self.data.columns[0]
            if yerr is None and 'std' in self.data.columns:
                yerr = 'std'
            # values = list(self.data[datafield])
            angles = deque(np.linspace(0, 2 * np.pi, self.data.shape[0] + 1)[:-1][::-1])
            angles.rotate(5)
            angles = list(angles)
            # for the values, repeat the first element to have a closed plot
            indices = list(range(self.data.shape[0])) + [0]
            y = list(self.data[datafield].iloc[indices])
            plt.plot(angles + [angles[0]], y, "k-", lw=3, **kwargs)
            if yerr:
                bounds0 = y - self.data[yerr].iloc[indices]
                plt.plot(angles + [angles[0]], bounds0, "k", lw=0.5, **kwargs)
                bounds1 = y + self.data[yerr].iloc[indices]
                plt.plot(angles + [angles[0]], bounds1, "k", lw=0.5, **kwargs)
            ax.set_xticks(angles)
            ax.set_xticklabels([_ for _ in self.data.index])
            ax.set_ylabel(self.unit)
            ax.set_title(
                "\n".join(wrap(f"{self.modality} anchored at {self.anchor._regionspec}", wrapwidth))
            )
            ax.tick_params(pad=9, labelsize=10)
            ax.tick_params(axis="y", labelsize=8)
            plt.tight_layout()
            return ax
        elif backend == "plotly":
            from plotly.express import line_polar
            df = pd.DataFrame(
                {
                    "values": pd.concat(
                        [
                            self.data["mean"],
                            self.data["mean"] - self.data["std"],
                            self.data["mean"] + self.data["std"]
                        ]
                    ),
                    "cat": (
                        len(self.data) * ["mean"]
                        + len(self.data) * ["mean - std"]
                        + len(self.data) * ["mean + std"]
                    )
                }
            )
            return line_polar(
                df, r="values", theta=df.index, color="cat", line_close=True, **kwargs
            )
        else:
            raise NotImplementedError

    def plot(
        self,
        *args,
        receptors: List[str] = None,
        backend: str = "matplotlib",
        **kwargs
    ):
        """
        Create a bar plot of receptor density fingerprint.

        Parameters
        ----------
        receptors : List[str], optional
            Plot a subset of receptors.
        backend: str
            "matplotlib", "plotly", or others supported by pandas DataFrame
            plotting backend.
        **kwargs
            takes Matplotlib.pyplot keyword arguments
        """
        kwargs['xlabel'] = ""
        kwargs["backend"] = backend
        og_data = self.data
        if receptors is not None:
            self._data_cached = og_data.loc[receptors]
        fig = super().plot(*args, **kwargs)
        self._data_cached = og_data
        return fig



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO
import os
from typing import Union, Dict, Tuple
import json

import numpy as np
import nibabel as nib
from neuroglancer_scripts.precomputed_io import get_IO_for_existing_dataset, PrecomputedIO
from neuroglancer_scripts.accessor import DataAccessError
from neuroglancer_scripts.http_accessor import HttpAccessor
from neuroglancer_scripts.mesh import read_precomputed_mesh, affine_transform_mesh
from nilearn.image.resampling import BoundingBoxError

from . import provider as _provider
from ...retrieval import requests, cache
from ...locations import boundingbox as _boundingbox
from ...commons import (
    logger,
    MapType,
    merge_meshes,
    SIIBRA_MAX_FETCH_SIZE_BYTES,
    SIIBRA_NG_VOL_USE_CACHE,
    QUIET,
    resample_img_to_img
)


def shift_ng_transfrom(
    transform_nm: np.ndarray, scale_resolution_nm: np.ndarray, max_resolution_nm: np.ndarray
) -> np.ndarray:
    """
    Helper method to get nifti standard affine.

    transfrorm.json stored with neuroglancer precomputed images and meshes
    are meant to be used for neuroglancer viewers and hence they are not
    representative of the affine in other tools. This method shifts back
    half a voxel in each axis.
    (see https://neuroglancer-scripts.readthedocs.io/en/latest/neuroglancer-info.html#different-conventions-for-coordinate-transformations)

    Parameters
    ----------
    transform_nm: np.ndarray
        Transform array created for displaying an image correctly from
        neuroglancer precomputed format in neuroglancer viewer.
    max_resolution_nm: np.ndarray
        The voxel resolution of the highest level of resolution.

    Returns
    -------
    np.ndarray
        Standard affine in nm
    """
    scaling = np.diag(np.r_[scale_resolution_nm, 1.0])
    affine = np.dot(transform_nm, scaling)
    affine[:3, 3] += (max_resolution_nm * 0.5)
    return affine


class NeuroglancerProvider(_provider.VolumeProvider, srctype="neuroglancer/precomputed"):

    def __init__(self, url: Union[str, Dict[str, str]]):
        _provider.VolumeProvider.__init__(self)
        self._init_url = url
        # TODO duplicated code to giftimesh
        if isinstance(url, str):  # one single image to load
            self._fragments = {None: NeuroglancerVolume(url)}
        elif isinstance(url, dict):  # assuming multiple for fragment images
            self._fragments = {n: NeuroglancerVolume(u) for n, u in url.items()}
        else:
            raise ValueError(f"Invalid url specified for {self.__class__.__name__}: {url}")

    @property
    def _url(self) -> Union[str, Dict[str, str]]:
        return self._init_url

    def fetch(
        self,
        fragment: str = None,
        resolution_mm: Union[float, Tuple[float, float, float]] = None,
        voi: _boundingbox.BoundingBox = None,
        max_bytes: float = SIIBRA_MAX_FETCH_SIZE_BYTES,
        **kwargs,
    ) -> nib.Nifti1Image:
        """
        Fetch 3D image data from neuroglancer volume.

        Parameters
        ----------
        fragment: str, optional
            The name of a fragment volume to fetch, if any. For example,
            some volumes are split into left and right hemisphere fragments.
            See :func:`~siibra.volumes.Volume.fragments`
        resolution_mm: float, default: None (i.e, the highest possible given max_bytes)
            Desired resolution in millimeters.
        voi: BoundingBox
            optional specification of a volume of interest to fetch.
        max_bytes: float: Default: NeuroglancerVolume.MAX_BYTES, 0.2GiB
            Maximum allowable size (in bytes) for downloading the image. siibra
            will attempt to find the highest resolution image with a size less
            than this value.
        """

        result = None

        if 'index' in kwargs:
            index = kwargs.pop('index')
            if fragment is not None:
                assert fragment == index.fragment
            fragment = index.fragment

        if len(self._fragments) > 1:
            if fragment is None:
                logger.info(
                    f"Merging fragments [{', '.join(self._fragments.keys())}]. "
                    f"You can select one using `fragment` kwarg."
                )
                result = self._merge_fragments(
                    resolution_mm=resolution_mm,
                    voi=voi,
                    max_bytes=max_bytes,
                    **kwargs
                )
            else:
                matched_names = [n for n in self._fragments if fragment.lower() in n.lower()]
                if len(matched_names) != 1:
                    raise ValueError(
                        f"Requested fragment '{fragment}' could not be matched uniquely "
                        f"to [{', '.join(self._fragments)}]"
                    )
                else:
                    result = self._fragments[matched_names[0]].fetch(
                        resolution_mm=resolution_mm, voi=voi, **kwargs
                    )
        else:
            assert len(self._fragments) > 0
            fragment_name, ngvol = next(iter(self._fragments.items()))
            if fragment is not None:
                assert fragment.lower() in fragment_name.lower()
            result = ngvol.fetch(
                resolution_mm=resolution_mm,
                voi=voi,
                max_bytes=max_bytes,
                **kwargs
            )

        # if a label is specified, mask the resulting image.
        if result is not None:
            if 'label' in kwargs:
                label = kwargs['label']
            elif ('index') in kwargs:
                label = kwargs['index'].label
            else:
                label = None
            if label is not None:
                result = nib.Nifti1Image(
                    (np.asanyarray(result.dataobj) == label).astype('uint8'),
                    result.affine,
                    dtype='uint8',
                )
            # result.set_qform(result.affine)  # TODO: needs to be investigated further

        return result

    def get_boundingbox(self, **fetch_kwargs) -> "_boundingbox.BoundingBox":
        """
        Return the bounding box in physical coordinates of the union of
        fragments in this neuroglancer volume.

        Parameters
        ----------
        fetch_kwargs:
            key word arguments that are used for fetching volumes,
            such as voi or resolution_mm.
        """
        bbox = None
        for frag in self._fragments.values():
            if len(frag.shape) > 3:
                logger.warning(
                    f"N-D Neuroglancer volume has shape {frag.shape}, but "
                    f"bounding box considers only {frag.shape[:3]}"
                )
            resolution_mm = fetch_kwargs.get("resolution_mm")
            if resolution_mm is None:
                affine = frag.affine
                shape = frag.shape[:3]
            else:
                scale = frag._select_scale(resolution_mm=resolution_mm)
                affine = scale.affine
                shape = scale.size[:3]
            next_bbox = _boundingbox.BoundingBox(
                (0, 0, 0), shape, space=None
            ).transform(affine)
            bbox = next_bbox if bbox is None else bbox.union(next_bbox)
        return bbox

    def _merge_fragments(
        self,
        resolution_mm: Union[float, Tuple[float, float, float]] = -1,
        voi: _boundingbox.BoundingBox = None,
        max_bytes: float = SIIBRA_MAX_FETCH_SIZE_BYTES,
    ) -> nib.Nifti1Image:
        with QUIET:
            if voi is not None:
                bbox = voi
            else:
                bbox = self.get_boundingbox(clip=False, background=0, resolution_mm=resolution_mm)

        num_conflicts = 0
        result = None
        for frag_name, frag_vol in self._fragments.items():
            frag_scale = frag_vol._select_scale(
                resolution_mm=resolution_mm,
                bbox=voi,
                max_bytes=max_bytes,
            )
            img = frag_scale.fetch(voi=voi)
            if img is None:
                logger.debug(f"Fragment {frag_name} did not provide content for fetching.")
                continue
            if result is None:
                # build the empty result image with its own affine and voxel space
                transl = np.identity(4)
                transl[:3, -1] = list(bbox.minpoint.transform(np.linalg.inv(img.affine)))
                result_affine = np.dot(img.affine, transl)  # adjust global bounding box offset to get global affine
                voxdims = np.asanyarray(np.ceil(
                    bbox.transform(np.linalg.inv(result_affine)).shape  # transform to the voxel space
                ), dtype="int")
                result_arr = np.zeros(voxdims, dtype=img.dataobj.dtype)
                result = nib.Nifti1Image(dataobj=result_arr, affine=result_affine)

            # resample to merge template and update it
            try:
                resampled_img = resample_img_to_img(source_img=img, target_img=result)
            except BoundingBoxError:
                logger.debug(f"Bounding box outside the fragment {frag_name}.")
                continue
            arr = np.asanyarray(resampled_img.dataobj)
            nonzero_voxels = arr != 0
            num_conflicts += np.count_nonzero(result_arr[nonzero_voxels])
            result_arr[nonzero_voxels] = arr[nonzero_voxels]

        if num_conflicts > 0:
            num_voxels = np.count_nonzero(result_arr)
            logger.warning(
                f"Merging fragments required to overwrite {num_conflicts} "
                f"conflicting voxels ({num_conflicts / num_voxels * 100.:2.3f}%)."
            )

        return result


class NeuroglancerVolume:

    USE_CACHE = SIIBRA_NG_VOL_USE_CACHE  # Whether to keep fetched data in local cache
    MAX_BYTES = SIIBRA_MAX_FETCH_SIZE_BYTES  # Number of bytes at which an image array is considered to large to fetch

    def __init__(self, url: str):
        assert isinstance(url, str)
        self.url = url
        self._scales_cached = None
        self._info = None
        self._transform_nm = None
        self._io: PrecomputedIO = None

    @property
    def transform_nm(self) -> np.ndarray:
        """
        This is the transformation matrix created to cater neuroglancer viewer
        for a neuroglancer precomputed images.
        """
        if self._transform_nm is not None:
            return self._transform_nm
        try:
            res = requests.HttpRequest(f"{self.url}/transform.json").get()
        except requests.SiibraHttpRequestError:
            res = None
        if res is not None:
            self._transform_nm = np.array(res)
            return self._transform_nm

        self._transform_nm = np.identity(4)
        logger.warning(f"No transform.json found at {self.url}, using identity.")
        return self._transform_nm

    @transform_nm.setter
    def transform_nm(self, val):
        self._transform_nm = val

    @property
    def io(self) -> PrecomputedIO:
        if self._io is None:
            accessor = HttpAccessor(self.url)
            self._io = get_IO_for_existing_dataset(accessor)
        return self._io

    @property
    def map_type(self):
        if self._info is None:
            self._bootstrap()
        return (
            MapType.LABELLED
            if self._info.get("type") == "segmentation"
            else MapType.STATISTICAL
        )

    @map_type.setter
    def map_type(self, val):
        if val is not None:
            logger.debug(
                "NeuroglancerVolume can determine its own maptype from self._info.get('type')"
            )

    def _bootstrap(self):
        self._info = requests.HttpRequest(f"{self.url}/info", func=lambda b: json.loads(b.decode())).get()
        self._scales_cached = sorted(
            [NeuroglancerScale(self, i) for i in self._info["scales"]]
        )

    @property
    def dtype(self):
        if self._info is None:
            self._bootstrap()
        return np.dtype(self._info["data_type"])

    @property
    def scales(self):
        if self._scales_cached is None:
            self._bootstrap()
        return self._scales_cached

    @property
    def shape(self):
        # return the shape of the scale 0 array
        return self.scales[0].size

    @property
    def affine(self):
        # return the affine matrix of the scale 0 data
        return self.scales[0].affine

    def fetch(
        self,
        resolution_mm: Union[float, Tuple[float, float, float]] = -1,
        voi: _boundingbox.BoundingBox = None,
        max_bytes: float = MAX_BYTES,
        **kwargs
    ):
        # the caller has to make sure voi is defined in the correct reference space
        scale = self._select_scale(resolution_mm=resolution_mm, bbox=voi, max_bytes=max_bytes)
        return scale.fetch(voi=voi, **kwargs)

    def get_shape(self, resolution_mm=None, max_bytes: float = MAX_BYTES):
        scale = self._select_scale(resolution_mm=resolution_mm, max_bytes=max_bytes)
        return scale.size

    def is_float(self):
        return self.dtype.kind == "f"

    def _select_scale(
        self,
        resolution_mm: Union[None, float, Tuple[float, float, float]],
        max_bytes: float = MAX_BYTES,
        bbox: _boundingbox.BoundingBox = None
    ) -> 'NeuroglancerScale':
        """
        Select the best possible scale for a given resolution and max_bytes constraint.

        If a specific resolution is requested, the smallest resolution <= max
        requested resolution is chosen.

        Parameters
        ----------
        resolution_mm (float/tuple): Requested resolution in mm. If negative, the highest resolution will be used.
        max_bytes (float): Maximum allowable memory bandwidth for the fetched image.

        Returns
        _______
            NeuroglancerScale: The best resolution scale available.
        """
        if resolution_mm is None or (isinstance(resolution_mm, (int, float)) and resolution_mm < 0):
            chosen_scale = min(self.scales)
        else:
            suitable_scales = sorted([s for s in self.scales if s.resolves(resolution_mm)])
            if not suitable_scales:
                raise ValueError(
                    f"Requested resolution {resolution_mm} is not feasible. Available resolutions: {[s.res_mm for s in self.scales]}"
                )
            chosen_scale = max(suitable_scales)  # `max` for choosing the closest resolution
            if not all('{:.6f}'.format(r).startswith(str(resolution_mm)) for r in chosen_scale.res_mm):
                logger.info(f"Closest resolution to requested is {chosen_scale.res_mm}")

        # Further reduce resolution to fit into max_bytes if needed.
        scale_changed = False
        while chosen_scale is not None and chosen_scale._estimate_nbytes(bbox) > max_bytes:
            next_scale = chosen_scale.next()
            if not next_scale:
                raise RuntimeError(
                    f"Infeasible resolution within {max_bytes / 1024**3} GiB."
                )
            chosen_scale = next_scale
            scale_changed = True

        if scale_changed:
            logger.warning((
                f"Changed resolution from {chosen_scale.res_mm} to {chosen_scale.res_mm} to allowed download size. ",
                "Increase `max_bytes` to allow higher-resolution downloads."
            ))

        return chosen_scale


class NeuroglancerScale:
    """One scale of a NeuroglancerVolume."""

    color_warning_issued = False

    def __init__(self, volume: NeuroglancerVolume, scaleinfo: dict):
        self.volume = volume
        self.chunk_sizes = np.array(scaleinfo["chunk_sizes"]).squeeze()
        self.encoding = scaleinfo["encoding"]
        self.key = scaleinfo["key"]
        self.res_nm = np.array(scaleinfo["resolution"]).squeeze()
        self.size = scaleinfo["size"]
        self.voxel_offset = np.array(scaleinfo["voxel_offset"])

    @property
    def res_mm(self):
        return self.res_nm / 1e6

    def resolves(self, resolution_mm: Union[float, Tuple[float, float, float]]):
        """Test whether the resolution of this scale is sufficient to provide the given resolution."""
        if isinstance(resolution_mm, tuple):
            return all(self.res_nm / 1e6 <= resolution_mm)
        if resolution_mm < 0:  # preserve legacy behaviour
            return True
        return all(r / 1e6 <= resolution_mm for r in self.res_nm)

    def __lt__(self, other: "NeuroglancerScale"):
        """Sort scales by resolution."""
        return all(self.res_nm[i] < other.res_nm[i] for i in range(3))

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"{self.__class__.__name__} {self.key}"

    def _estimate_nbytes(self, bbox: _boundingbox.BoundingBox = None):
        """Estimate the size image array to be fetched in bytes, given a bounding box."""
        if bbox is None:
            bbox_ = _boundingbox.BoundingBox((0, 0, 0), self.size, space=None)
        else:
            bbox_ = bbox.transform(np.linalg.inv(self.affine))
        result = self.volume.dtype.itemsize * bbox_.volume
        logger.debug(
            f"Approximate size for fetching resolution "
            f"({', '.join(map('{:.6f}'.format, self.res_mm))}) mm "
            f"is {result / 1024**3:.5f} GiB."
        )
        return result

    def next(self):
        """Returns the next scale in this volume, of None if this is the last."""
        my_index = self.volume.scales.index(self)
        if my_index < len(self.volume.scales):
            return self.volume.scales[my_index + 1]
        else:
            return None

    def prev(self):
        """Returns the previous scale in this volume, or None if this is the first."""
        my_index = self.volume.scales.index(self)
        print(f"Index of {self.key} is {my_index} of {len(self.volume.scales)}.")
        if my_index > 0:
            return self.volume.scales[my_index - 1]
        else:
            return None

    @property
    def affine(self):
        affine_ = shift_ng_transfrom(
            transform_nm=self.volume.transform_nm,
            scale_resolution_nm=self.res_nm,
            max_resolution_nm=self.volume.scales[0].res_nm[0],
        )
        affine_[:3, :] /= 1e6
        return affine_

    def _point_to_lower_chunk_idx(self, xyz):
        return (
            np.floor((np.array(xyz) - self.voxel_offset) / self.chunk_sizes)
            .astype("int")
            .ravel()
        )

    def _point_to_upper_chunk_idx(self, xyz):
        return (
            np.ceil((np.array(xyz) - self.voxel_offset) / self.chunk_sizes)
            .astype("int")
            .ravel()
        )

    def _read_chunk(self, gx, gy, gz, channel: int = None):
        if any(v < 0 for v in (gx, gy, gz)):
            raise DataAccessError('Negative tile index observed - you have likely requested fetch() with a voi specification ranging outside the actual data.')
        if self.volume.USE_CACHE:
            cachefile = cache.CACHE.build_filename(
                "{}_{}_{}_{}_{}".format(self.volume.url, self.key, gx, gy, gz),
                suffix=".npy",
            )
            if os.path.isfile(cachefile):
                return np.load(cachefile)

        x0 = gx * self.chunk_sizes[0]
        y0 = gy * self.chunk_sizes[1]
        z0 = gz * self.chunk_sizes[2]
        x1, y1, z1 = np.minimum(self.chunk_sizes + [x0, y0, z0], self.size)
        chunk_czyx = self.volume.io.read_chunk(self.key, (x0, x1, y0, y1, z0, z1))
        if channel is None:
            channel = 0
            if chunk_czyx.shape[0] > 1 and not self.color_warning_issued:
                logger.warning(
                    f"The volume has {chunk_czyx.shape[0]} color channels. "
                    "Returning the first channel now but you can specify one "
                    "with 'channel' keyword."
                )
                self.color_warning_issued = True
        elif channel + 1 > chunk_czyx.shape[0]:
            raise ValueError(f"There are only {chunk_czyx.shape[0]} color channels.")
        chunk_zyx = chunk_czyx[channel]

        if self.volume.USE_CACHE:
            np.save(cachefile, chunk_zyx)
        return chunk_zyx

    def fetch(self, voi: _boundingbox.BoundingBox = None, **kwargs):

        # define the bounding box in this scale's voxel space
        if voi is None:
            bbox_ = _boundingbox.BoundingBox((0, 0, 0), self.size, space=None)
        else:
            bbox_ = voi.transform(np.linalg.inv(self.affine), space=None)

        # extract minimum and maximum the chunk indices to be loaded
        gx0, gy0, gz0 = self._point_to_lower_chunk_idx(tuple(bbox_.minpoint))
        gx1, gy1, gz1 = self._point_to_upper_chunk_idx(tuple(bbox_.maxpoint))

        # create requested data volume, and fill it with the required chunk data
        shape_zyx = np.array([gz1 - gz0, gy1 - gy0, gx1 - gx0]) * self.chunk_sizes[::-1]
        data_zyx = None
        for gx in range(gx0, gx1):
            x0 = (gx - gx0) * self.chunk_sizes[0]
            for gy in range(gy0, gy1):
                y0 = (gy - gy0) * self.chunk_sizes[1]
                for gz in range(gz0, gz1):
                    try:
                        chunk = self._read_chunk(gx, gy, gz, kwargs.get("channel"))
                    except DataAccessError:
                        logger.debug(f"voi: {voi}", exc_info=1)
                        continue
                    if data_zyx is None:
                        data_zyx = np.zeros(shape_zyx, dtype=self.volume.dtype)
                    z0 = (gz - gz0) * self.chunk_sizes[2]
                    z1, y1, x1 = np.array([z0, y0, x0]) + chunk.shape
                    data_zyx[z0:z1, y0:y1, x0:x1] = chunk

        # no voxel values in voi
        if data_zyx is None:
            return None

        # determine the remaining offset from the "chunk mosaic" to the
        # exact bounding box requested, to cut off undesired borders
        data_min = np.array([gx0, gy0, gz0]) * self.chunk_sizes
        x0, y0, z0 = (np.array(bbox_.minpoint) - data_min).astype("int")
        xd, yd, zd = np.ceil(bbox_.maxpoint).astype(int) - np.floor(bbox_.minpoint).astype(int)
        offset = tuple(bbox_.minpoint)
        if voi is not None:
            logger.debug(
                f"Input: {voi.minpoint.coordinate}, {voi.maxpoint.coordinate}.\nVoxel space: {bbox_.minpoint.coordinate}, {bbox_.maxpoint.coordinate}"
            )

        # build the nifti image
        trans = np.identity(4)[[2, 1, 0, 3], :]  # zyx -> xyz
        shift = np.c_[np.identity(4)[:, :3], np.r_[offset, 1]]
        return nib.Nifti1Image(
            data_zyx[z0: z0 + zd, y0: y0 + yd, x0: x0 + xd],
            np.dot(self.affine, np.dot(shift, trans)),
        )


class NeuroglancerMesh(_provider.VolumeProvider, srctype="neuroglancer/precompmesh"):
    """
    A surface mesh provided as neuroglancer precomputed mesh.
    """

    @staticmethod
    def _fragmentinfo(url: str) -> Dict[str, Union[str, np.ndarray, Dict]]:
        """Prepare basic mesh fragment information from url."""
        return {
            "url": url,
            "transform_nm": np.array(requests.HttpRequest(f"{url}/transform.json").data),
            "info": requests.HttpRequest(url=f"{url}/info", func=requests.DECODERS['.json']).data
        }

    # TODO check resource typing?
    def __init__(self, resource: Union[str, dict], volume=None):
        self.volume = volume
        self._init_url = resource
        if isinstance(resource, str):
            self._meshes = {None: self._fragmentinfo(resource)}
        elif isinstance(resource, dict):
            self._meshes = {n: self._fragmentinfo(u) for n, u in resource.items()}
        else:
            raise ValueError(f"Resource specification not understood for {self.__class__.__name__}: {resource}")

    @property
    def _url(self) -> Union[str, Dict[str, str]]:
        return self._init_url

    def get_boundingbox(self, clip=False, background=0.0, **fetch_kwargs) -> '_boundingbox.BoundingBox':
        """
        Bounding box calculation is not yet implemented for meshes.
        """
        raise NotImplementedError(
            f"Bounding box access to {self.__class__.__name__} objects not yet implemented."
        )

    def _get_fragment_info(self, meshindex: int) -> Dict[str, Tuple[str, ]]:
        # extract available fragment urls with their names for the given mesh index
        result = {}

        for name, spec in self._meshes.items():
            mesh_key = spec.get('info', {}).get('mesh')
            meshurl = f"{spec['url']}/{mesh_key}/{str(meshindex)}:0"
            transform = spec.get('transform_nm')
            try:
                meshinfo = requests.HttpRequest(url=meshurl, func=requests.DECODERS['.json']).data
            except requests.SiibraHttpRequestError:
                continue
            fragment_names = meshinfo.get('fragments')

            if len(fragment_names) == 0:
                raise RuntimeError(f"No fragments found at {meshurl}")
            elif len(self._meshes) > 1:
                # multiple meshes were configured, so we expect only one fragment under each mesh url
                if len(fragment_names) > 1:
                    raise RuntimeError(
                        f"{self.__class__.__name__} was configured with multiple mesh fragments "
                        f"({', '.join(self._meshes.keys())}), but unexpectedly even more fragmentations "
                        f"were found at {spec['url']}."
                    )
                result[name] = (f"{spec['url']}/{mesh_key}/{fragment_names[0]}", transform)
            else:
                # only one mesh was configures, so we might still
                # see multiple fragments under the mesh url
                for fragment_name in fragment_names:
                    result[fragment_name] = (f"{spec['url']}/{mesh_key}/{fragment_name}", transform)

        return result

    def _fetch_fragment(self, url: str, transform_nm: np.ndarray):
        r = requests.HttpRequest(url, func=lambda b: BytesIO(b))
        (vertices_vox, triangles_vox) = read_precomputed_mesh(r.data)
        vertices, triangles = affine_transform_mesh(vertices_vox, triangles_vox, transform_nm)
        vertices /= 1e6
        return {'verts': vertices, 'faces': triangles}

    def fetch(self, label: int, fragment: str):
        """
        Fetches a particular mesh. Each mesh is a dictionary with keys:

        Parameters
        ----------
        label: int
            Label of the volume
        fragment: str, default: None
            A fragment name can be specified to choose from multiple fragments.

            Note
            ----
            If not specified, multiple fragments will be merged into one mesh.
            In such a case, the verts and faces arrays of different fragments
            are appended to one another.
        Returns
        -------
        dict
            - 'verts': An Nx3 array of vertex coordinates (in nanometer)
            - 'faces': an MX3 array containing connection data of vertices
            - 'name': Name of the of the mesh variant
        """

        # extract fragment information for the requested mesh
        fragment_infos = self._get_fragment_info(label)

        if fragment is None:

            # no fragment specified, return merged fragment meshes
            if len(fragment_infos) == 1:
                url, transform = next(iter(fragment_infos.values()))
                return self._fetch_fragment(url, transform)
            else:
                logger.info(
                    f"Fragments [{', '.join(fragment_infos.keys())}] are merged during fetch(). "
                    "You can select one of them using the 'fragment' parameter."
                )
                return merge_meshes([self._fetch_fragment(u, t) for u, t in fragment_infos.values()])

        else:

            # match fragment to available fragments
            matched = [
                info for name, info in fragment_infos.items()
                if fragment.lower() in name
            ]
            if len(matched) == 1:
                url, transform = next(iter(matched))
                return self._fetch_fragment(url, transform)
            else:
                raise ValueError(
                    f"The requested mesh fragment name '{fragment}' could not be resolved. "
                    f"Valid names are: {', '.join(fragment_infos.keys())}"
                )


class NeuroglancerSurfaceMesh(NeuroglancerMesh, srctype="neuroglancer/precompmesh/surface"):
    """
    Only shadows NeuroglancerMesh for the special surface srctype,
    which provides a mesh urls plus a mesh index for identifying the surface.
    Behaves like NeuroglancerMesh otherwise.

    TODO this class might be replaced by implementing a default label index for the parent class.
    """
    def __init__(self, spec: str, **kwargs):
        # Here we expect a string of the form "<url> <labelindex>",
        # and use this to set the url and label index in the parent class.
        assert ' ' in spec
        url, labelindex, *args = spec.split(' ')
        assert labelindex.isnumeric()
        self.label = int(labelindex)
        NeuroglancerMesh.__init__(self, resource=url, **kwargs)

    @property
    def fragments(self, meshindex=1):
        """
        Returns the set of fragment names available
        for the mesh with the given index.
        """
        return set(self._get_fragment_info(self.label))

    def fetch(self, **kwargs):
        if 'fragment' not in kwargs:
            kwargs['fragment'] = None
        return super().fetch(label=self.label, **kwargs)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Union, Dict, List
import json

from nibabel import Nifti1Image

if TYPE_CHECKING:
    from ...locations.boundingbox import BoundingBox

VolumeData = Union[Nifti1Image, Dict]


class VolumeProvider(ABC):

    _SUBCLASSES: List[VolumeProvider] = []

    def __init_subclass__(cls, srctype: str) -> None:
        cls.srctype = srctype
        VolumeProvider._SUBCLASSES.append(cls)
        return super().__init_subclass__()

    @abstractmethod
    def get_boundingbox(self, clip=True, background=0.0) -> "BoundingBox":
        raise NotImplementedError

    @property
    def fragments(self) -> List[str]:
        return []

    @abstractmethod
    def fetch(self, *args, **kwargs) -> VolumeData:
        raise NotImplementedError

    @property
    @abstractmethod
    def _url(self) -> Union[str, Dict[str, str]]:
        """
        This is needed to provide urls to applications that can utilise such resources directly.
        e.g. siibra-api
        """
        return {}


class SubvolumeProvider(VolumeProvider, srctype="subvolume"):
    """
    This provider wraps around an existing volume provider,
    but is preconfigured to always fetch a fixed subvolume.
    The primary use is to provide a fixed z coordinate
    of a 4D volume provider as a 3D volume under the
    interface of a normal volume provider.
    """

    _USE_CACHING = False
    _FETCHED_VOLUMES = {}

    class UseCaching:
        def __enter__(self):
            SubvolumeProvider._USE_CACHING = True

        def __exit__(self, et, ev, tb):
            SubvolumeProvider._USE_CACHING = False
            SubvolumeProvider._FETCHED_VOLUMES = {}

    def __init__(self, parent_provider: VolumeProvider, z: int):
        VolumeProvider.__init__(self)
        self.provider = parent_provider
        self.srctype = parent_provider.srctype
        self.z = z

    def get_boundingbox(self, clip=True, background=0.0, **fetch_kwargs) -> "BoundingBox":
        return self.provider.get_boundingbox(clip=clip, background=background, **fetch_kwargs)

    def fetch(self, **kwargs):
        # activate caching at the caller using "with SubvolumeProvider.UseCaching():""
        if self.__class__._USE_CACHING:
            data_key = json.dumps(self.provider._url, sort_keys=True) \
                + json.dumps(kwargs, sort_keys=True)
            if data_key not in self.__class__._FETCHED_VOLUMES:
                vol = self.provider.fetch(**kwargs)
                self.__class__._FETCHED_VOLUMES[data_key] = vol
            vol = self.__class__._FETCHED_VOLUMES[data_key]
        else:
            vol = self.provider.fetch(**kwargs)
        return vol.slicer[:, :, :, self.z]

    def __getattr__(self, attr):
        return self.provider.__getattribute__(attr)

    @property
    def _url(self) -> Union[str, Dict[str, str]]:
        return super()._url



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, Dict, Tuple, List

from textwrap import wrap
import numpy as np
import pandas as pd

from . import tabular
from .. import anchor as _anchor
from ..feature import Compoundable


class CorticalProfile(tabular.Tabular, Compoundable):
    """
    Represents a 1-dimensional profile of measurements along cortical depth,
    measured at relative depths between 0 representing the pial surface,
    and 1 corresponding to the gray/white matter boundary.

    Mandatory attributes are the list of depth coordinates and the list of
    corresponding measurement values, which have to be of equal length,
    as well as a unit and description of the measurements.

    Optionally, the depth coordinates of layer boundaries can be specified.

    Most attributes are modelled as properties, so dervide classes are able
    to implement lazy loading instead of direct initialization.

    """

    LAYERS = {0: "0", 1: "I", 2: "II", 3: "III", 4: "IV", 5: "V", 6: "VI", 7: "WM"}
    BOUNDARIES = list(zip(list(LAYERS.keys())[:-1], list(LAYERS.keys())[1:]))

    _filter_attrs = ["modality"]
    _compound_attrs = ["modality"]

    def __init__(
        self,
        description: str,
        modality: str,
        anchor: _anchor.AnatomicalAnchor,
        depths: Union[list, np.ndarray] = None,
        values: Union[list, np.ndarray] = None,
        unit: str = None,
        boundary_positions: Dict[Tuple[int, int], float] = None,
        datasets: list = [],
        id: str = None,
        prerelease: bool = False,
    ):
        """Initialize profile.

        Parameters
        ----------
        description: str
            Human-readable of the modality of the measurements.
        modality: str
            Short textual description of the modality of measurements.
        anchor: AnatomicalAnchor
        depths: list, default: None
            List of cortical depth positions corresponding to each
            measurement, all in the range [0..1]
        values: list, default: None
            List of the actual measurements at each depth position.
            Length must correspond to 'depths'.
        unit: str, default: None
            Textual identifier for the unit of measurements.
        boundary_positions: dict, default: None
            Dictionary of depths at which layer boundaries were identified.
            Keys are tuples of layer numbers, e.g. (1,2), and values are
            cortical depth positions in the range [0..1].
        datasets : list[Dataset]
            list of datasets corresponding to this feature
        """

        # cached properties will be revealed as property functions,
        # so derived classes may choose to override for lazy loading.
        self._unit = unit
        self._depths_cached = depths
        self._values_cached = values
        self._boundary_positions = boundary_positions

        tabular.Tabular.__init__(
            self,
            modality=modality,
            description=description,
            anchor=anchor,
            data=None,  # lazy loader below
            datasets=datasets,
            id=id,
            prerelease=prerelease,
        )

    def _check_sanity(self):
        # check plausibility of the profile
        assert isinstance(self._depths, (list, np.ndarray)), "Some depths are not valid"
        assert isinstance(self._values, (list, np.ndarray)), "Some values are not valid"
        assert len(self._values) == len(self._depths), "There exist uneven number of depths and values"
        assert all(0 <= d <= 1 for d in self._depths), "Some depths is not between 0 and 1"

        if self.boundaries_mapped:
            try:
                assert all(0 <= d <= 1 for d in self.boundary_positions.values()), "Some boundary positions are not between 0 and 1"
            except AssertionError:
                assert np.allclose([d for d in self.boundary_positions.values() if d > 1], 1, rtol=1e-15)
                assert np.allclose([d for d in self.boundary_positions.values() if d < 0], 0, rtol=1e-15)
            assert all(
                layerpair in self.BOUNDARIES
                for layerpair in self.boundary_positions.keys()
            ), "Some value in BOUNDARIES are not mapped in boundary_positions"

    @property
    def unit(self) -> str:
        """Optionally overridden in derived classes."""
        if self._unit is None:
            raise NotImplementedError(f"'unit' not set for {self.__class__.__name__}.")
        return self._unit

    @property
    def boundary_positions(self) -> Dict[Tuple[int, int], float]:
        if self._boundary_positions is None:
            return {}
        else:
            return self._boundary_positions

    def assign_layer(self, depth: float):
        """Compute the cortical layer for a given depth from the
        layer boundary positions. If no positions are available
        for this profile, return None."""
        assert 0 <= depth <= 1
        if len(self.boundary_positions) == 0:
            return None
        else:
            return max(
                [l2 for (l1, l2), d in self.boundary_positions.items() if d < depth]
            )

    @property
    def boundaries_mapped(self) -> bool:
        if self.boundary_positions is None:
            return False
        else:
            return all((b in self.boundary_positions) for b in self.BOUNDARIES)

    @property
    def _layers(self):
        """List of layers assigned to each measurements,
        if layer boundaries are available for this features.
        """
        if self.boundaries_mapped:
            return [self.assign_layer(d) for d in self._depths]
        else:
            return None

    @property
    def data(self):
        """Return a pandas Series representing the profile."""
        self._check_sanity()
        iscompound = len(self._values.shape) > 1 and self._values.shape[1] == 2
        if iscompound:
            columns = [f"{self.modality} mean ({self.unit})", "std"]
        else:
            columns = [f"{self.modality} ({self.unit})"]
        return pd.DataFrame(self._values, index=self._depths, columns=columns)

    @classmethod
    def _merge_elements(
        cls,
        elements: List["CorticalProfile"],
        description: str,
        modality: str,
        anchor: _anchor.AnatomicalAnchor,
    ):
        assert all(np.array_equal(elements[0]._depths, f._depths) for f in elements)
        assert len({f.unit for f in elements}) == 1
        values_stacked = np.stack([f._values for f in elements])
        return CorticalProfile(
            description=description,
            modality=modality,
            anchor=anchor,
            depths=np.stack([f._depths for f in elements]).mean(0),
            values=np.stack([values_stacked.mean(0), values_stacked.std(0)]).T,
            unit=elements[0].unit,
            boundary_positions=None,
        )

    def plot(self, *args, backend="matplotlib", **kwargs):
        """
        Plot the profile.

        Parameters
        ----------
        backend: str
            "matplotlib", "plotly", or others supported by pandas DataFrame
            plotting backend.
        **kwargs
            Keyword arguments are passed on to the plot command. 'layercolor'
            can be used to specify a color for cortical layer shading.
        """
        wrapwidth = kwargs.pop("textwrap") if "textwrap" in kwargs else 40
        kwargs["title"] = kwargs.get("title", "\n".join(wrap(self.name, wrapwidth)))
        layercolor = kwargs.pop("layercolor", "gray")

        iscompound = len(self._values.shape) > 1 and self._values.shape[1] == 2
        ymax = max(
            0,
            sum(self._values.max(axis=0)) if iscompound else self._values.max()
        )
        if backend == "matplotlib":
            kwargs["xlabel"] = kwargs.get("xlabel", "Cortical depth")
            kwargs["ylabel"] = kwargs.get("ylabel", self.unit)
            kwargs["grid"] = kwargs.get("grid", True)
            axs = self.data.iloc[:, 0].plot(*args, **kwargs, backend=backend)
            axs.set_ylim(kwargs.get("ylim", (0, ymax)))

            if self.boundaries_mapped:
                bvals = list(self.boundary_positions.values())
                for i, (d1, d2) in enumerate(list(zip(bvals[:-1], bvals[1:]))):
                    axs.text(
                        d1 + (d2 - d1) / 2.0,
                        10,
                        self.LAYERS[i + 1],
                        weight="normal",
                        ha="center",
                    )
                    if i % 2 == 0:
                        axs.axvspan(d1, d2, color=layercolor, alpha=0.3)

            axs.set_title(axs.get_title(), fontsize="medium")

            if iscompound:
                axs.set_ylabel(f"average {kwargs['ylabel']} \u00b1 std")
                av = self.data.values[:, 0]
                std = self.data.values[:, 1]
                axs.fill_between(self.data.index.values, av - std, av + std, alpha=0.5)

            return axs

        elif backend == "plotly":
            kwargs["title"] = kwargs["title"].replace("\n", "<br>")
            kwargs["labels"] = {
                "index": kwargs.pop("xlabel", None) or kwargs.pop("index", "Cortical depth"),
                "value": kwargs.pop("ylabel", None) or kwargs.pop("value", self.unit)
            }
            fig = self.data.iloc[:, 0].plot(*args, **kwargs, backend=backend)
            if self.boundaries_mapped:
                bvals = list(self.boundary_positions.values())
                for i, (d1, d2) in enumerate(list(zip(bvals[:-1], bvals[1:]))):
                    fig.add_vrect(
                        x0=d1, x1=d2, line_width=0, fillcolor=layercolor,
                        opacity=0.2 if i % 2 == 0 else 0.0,
                        label=dict(text=self.LAYERS[i + 1], textposition="bottom center")
                    )
            fig.update_layout(
                showlegend=False,
                yaxis_range=(0, ymax),
                title=dict(
                    automargin=True, yref="container", xref="container",
                    pad=dict(t=40), xanchor="left", yanchor="top"
                )
            )
            if iscompound:
                from plotly.graph_objects import Scatter
                x = self.data.index.values
                av = self.data.values[:, 0]
                std = self.data.values[:, 1]
                fig.update_layout(yaxis_title=f"average {kwargs['labels']['value']} &plusmn; std")
                fig.add_traces(
                    Scatter(
                        x=np.concatenate((x, x[::-1])),  # x, then x reversed
                        y=np.concatenate((av + std, (av - std)[::-1])),  # upper, then lower reversed
                        fill='toself',
                        fillcolor='rgba(0,100,80,0.5)',
                        line=dict(color='rgba(255,255,255,0)'),
                        hoverinfo="skip",
                        showlegend=False
                    )
                )
            return fig
        else:
            return self.data.plot(*args, **kwargs, backend=backend)

    @property
    def _depths(self):
        """
        Returns a list of the relative cortical depths of the measured values in the range [0..1].

        To be implemented in derived class.
        """
        if self._depths_cached is None:
            raise NotImplementedError(
                f"'_depths' not available for {self.__class__.__name__}."
            )
        return self._depths_cached

    @property
    def _values(self):
        """
        Returns a list of the measured values per depth.

        To be implemented in derived class.
        """
        if self._values_cached is None:
            raise NotImplementedError(
                f"'_values' not available for {self.__class__.__name__}."
            )
        return self._values_cached

    @property
    def name(self):
        if hasattr(self, "receptor"):
            return super().name + f": {self.receptor}"
        if hasattr(self, "location"):
            return super().name + f": {self.location.coordinate}"
        return super().name



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Query data features published as Ebrains datasets with AtlasConcepts"""

from collections import defaultdict
import re
from packaging.version import Version
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from . import query
from ..features.dataset import ebrains as _ebrains
from ..commons import logger, siibra_tqdm
from ..features import anchor as _anchor
from ..retrieval import requests, datasets, cache
from ..core.concept import get_registry

if TYPE_CHECKING:
    from ..core.region import Region


class EbrainsFeatureQuery(query.LiveQuery, args=[], FeatureType=_ebrains.EbrainsDataFeature):
    # in EBRAINS knowledge graph prior to v3, versions were modelled
    # in dataset names. Typically found formats are (v1.0) and [rat, v2.1]
    VERSION_PATTERN = re.compile(r"^(.*?) *[\[\(][^v]*?(v[0-9].*?)[\]\)]")
    COMPACT_FEATURE_LIST = True

    # datasets whose name contains any of these strings will be ignored
    _BLACKLIST = {
        "Whole-brain parcellation of the Julich-Brain Cytoarchitectonic Atlas",
        "whole-brain collections of cytoarchitectonic probabilistic maps",
        "DiFuMo atlas",
        "Automated Anatomical Labeling (AAL1) atlas",
    }

    loader = requests.MultiSourcedRequest(
        requests=[
            requests.GitlabProxy(
                flavour=requests.GitlabProxyEnum.PARCELLATIONREGION_V1,
            ),
        ]
    )

    parcellation_ids = None

    def __init__(self, **kwargs):
        query.LiveQuery.__init__(self, **kwargs)

        if self.__class__.parcellation_ids is None:
            self.__class__.parcellation_ids = [
                dset.id
                for parc in get_registry("Parcellation")
                for dset in parc.datasets
                if isinstance(dset, datasets.EbrainsV3DatasetVersion)
            ]

    def query(self, region: "Region"):
        versioned_datasets = defaultdict(dict)
        invalid_species_datasets = {}
        results = self.loader.data.get("results", [])

        for r in siibra_tqdm(results, total=len(results), unit='dataset', desc=f'Matching datasets to "{region}"'):
            regionname = r.get("name", None)
            alias = r.get("alias", None)
            for ds_spec in r.get("datasets", []):
                ds_name = ds_spec.get("name")
                ds_id = ds_spec.get("@id")
                if "dataset" not in ds_id:
                    continue

                try:
                    ds_species = _anchor.Species.decode(ds_spec)
                except ValueError:
                    logger.debug(f"Cannot decode {ds_spec}")
                    invalid_species_datasets[ds_id] = ds_name
                    continue

                if self.COMPACT_FEATURE_LIST:
                    if any(ds_id.endswith(i) for i in self.parcellation_ids):
                        continue
                    if any(e.lower() in ds_name.lower() for e in self._BLACKLIST):
                        continue

                dset = _ebrains.EbrainsDataFeature(
                    dataset_version_id=ds_id,
                    anchor=_anchor.AnatomicalAnchor(
                        region=alias or regionname,
                        species=ds_species,
                    ),
                )
                if not dset.matches(region):
                    continue

                version_match = self.VERSION_PATTERN.search(ds_name)
                if version_match is None or not self.COMPACT_FEATURE_LIST:
                    yield dset
                else:  # store version, add only the latest version after the loop
                    name, version = version_match.groups()
                    versioned_datasets[name][version] = dset

        if len(invalid_species_datasets) > 0:
            with NamedTemporaryFile(mode="w", suffix=".txt", delete=False, encoding="utf-8") as f:
                for dsid, dsname in invalid_species_datasets.items():
                    f.write(f"{dsid} {dsname}\n")
                logger.warning(
                    f"{len(invalid_species_datasets)} datasets have been ignored, "
                    "because siibra could not decode their species. "
                    f"See {f.name}"
                )

        # if versioned datasets have been recorded, register only
        # the newest one with older ones linked as a version history.
        for name, dsets in versioned_datasets.items():
            try:  # if possible, sort by version tag
                sorted_versions = sorted(dsets.keys(), key=Version)
            except TypeError:  # else sort lexicographically
                sorted_versions = sorted(dsets.keys())

            # chain the dataset versions
            prev = None
            for version in sorted_versions:
                curr = dsets[version]
                curr.version = version
                if prev is not None:
                    curr._prev = prev
                    prev._next = curr
                prev = curr

            logger.debug(
                f"Registered only version {version} of {', '.join(sorted_versions)} for {name}. "
                f"Its version history is: {curr.version_history}"
            )
            yield curr


cache.Warmup.register_warmup_fn(cache.WarmupLevel.DATA)(lambda: EbrainsFeatureQuery.loader.data)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Qualification between two BrainStructures"""

from enum import Enum
from dataclasses import dataclass
from typing import Dict, Generic, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from .structure import BrainStructure

T = TypeVar("T")


class Qualification(Enum):
    EXACT = 1
    APPROXIMATE = 2
    OTHER_VERSION = 3
    CONTAINED = 4
    CONTAINS = 5
    OVERLAPS = 6
    HOMOLOGOUS = 7

    @property
    def verb(self):
        """
        a string that can be used as a verb in a sentence
        for producing human-readable messages.
        """
        transl = {
            Qualification.EXACT: 'coincides with',
            Qualification.OVERLAPS: 'overlaps with',
            Qualification.CONTAINED: 'is contained in',
            Qualification.CONTAINS: 'contains',
            Qualification.APPROXIMATE: 'approximates to',
            Qualification.HOMOLOGOUS: 'is homologous to',
            Qualification.OTHER_VERSION: 'is another version of'
        }
        assert self in transl, f"{str(self)} verb cannot be found."
        return transl[self]

    def invert(self):
        """
        Return qualification with the inverse meaning
        """
        inverses = {
            Qualification.EXACT: Qualification.EXACT,
            Qualification.OVERLAPS: Qualification.OVERLAPS,
            Qualification.CONTAINED: Qualification.CONTAINS,
            Qualification.CONTAINS: Qualification.CONTAINED,
            Qualification.APPROXIMATE: Qualification.APPROXIMATE,
            Qualification.HOMOLOGOUS: Qualification.HOMOLOGOUS,
            Qualification.OTHER_VERSION: Qualification.OTHER_VERSION,
        }
        assert self in inverses, f"{str(self)} inverses cannot be found."
        return inverses[self]

    def __lt__(self, other: "Qualification"):
        """
        This is used to sort feature query results. Since it is very difficult
        to determine a well-ordering principle and it is difficult to sort
        without one, the enum values are used for sorting. This means
        not all comparisons have logical basis but they are well-defined,
        making it reproducible but also clearly distinguishes important
        comparisons.
        """
        return self.value < other.value

    def __str__(self):
        return f"{self.__class__.__name__}={self.name.lower()}"

    def __repr__(self):
        return str(self)

    @staticmethod
    def parse_relation_assessment(spec: Dict):
        name = spec.get("name")
        if name == "is homologous to":
            return Qualification.HOMOLOGOUS
        raise Exception(f"Cannot parse spec: {spec}")


@dataclass
class AnatomicalAssignment(Generic[T]):
    """Represents a qualified assignment between anatomical structures."""
    query_structure: "BrainStructure"
    assigned_structure: "BrainStructure"
    qualification: Qualification
    explanation: str = ""

    @property
    def is_exact(self):
        return self.qualification == Qualification.EXACT

    def __str__(self):
        msg = f"'{self.query_structure}' {self.qualification.verb} '{self.assigned_structure}'"
        return msg if self.explanation == "" else f"{msg} - {self.explanation}"

    def invert(self):
        return AnatomicalAssignment(
            self.assigned_structure,
            self.query_structure,
            self.qualification.invert(),
            self.explanation
        )

    def __lt__(self, other: 'AnatomicalAssignment'):
        if not isinstance(other, AnatomicalAssignment):
            raise ValueError(f"Cannot compare AnatomicalAssignment with instances of '{type(other)}'")
        return self.qualification < other.qualification



from setuptools import setup, find_packages
import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))


def find_version():
    path_to_version = os.path.join(ROOT_DIR, "siibra", "VERSION")
    with open(path_to_version, "r", encoding="utf-8") as f:
        return f.read()


with open(os.path.join(ROOT_DIR, "README.rst"), "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="siibra",
    version=find_version(),
    author="Big Data Analytics Group, Forschungszentrum Juelich, Institute of Neuroscience and Medicine (INM-1)",
    author_email="inm1-bda@fz-juelich.de",
    description="siibra - Software interfaces for interacting with brain atlases",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/FZJ-INM1-BDA/siibra-python",
    packages=find_packages(include=["siibra", "siibra.*"]),
    include_package_data=True,
    package_data={
        'siibra': [
            'VERSION',
            'vocabularies/gene_names.json',
            'vocabularies/receptor_symbols.json',
            'vocabularies/region_aliases.json'
        ]
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
    python_requires=">=3.7",
    install_requires=[
        "anytree",
        "nibabel",
        "appdirs",
        "scikit-image",
        "requests",
        "neuroglancer-scripts",
        "nilearn",
        'typing-extensions; python_version < "3.8"',
        "filelock",
        "ebrains-drive >= 0.6.0",
    ],
)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Package handling variety of volumes and volume operations"""

from typing import List, Union

import numpy as np

from .parcellationmap import Map
from .providers import provider
from .volume import from_array, from_file, from_pointcloud, from_nifti, Volume
from ..commons import logger



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represents lists of probabilistic brain region maps."""

from os import path, makedirs
from typing import Dict, List

import numpy as np

from . import parcellationmap, volume as _volume
from .providers import provider
from ..commons import MapIndex, logger, connected_components, siibra_tqdm, resample_img_to_img
from ..locations import boundingbox
from ..retrieval.cache import CACHE
from ..retrieval.requests import HttpRequest, FileLoader


class SparseIndex:

    # Precomputed sparse indices are stored in an EBRAINS data proxy
    _BUCKET = "https://data-proxy.ebrains.eu/api/v1/buckets/reference-atlas-data/"
    _DATAPROXY_BASEURL = _BUCKET + "sparse-indices/sparse-indices-siibra_python_v1.0/"

    _SUFFIXES = {
        "probs": ".sparseindex_v1.probs.txt.gz",
        "bboxes": ".sparseindex_v1.bboxes.txt.gz",
        "voxels": ".sparseindex_v1.voxels.nii.gz"
    }

    def __init__(self):
        self.probs = []
        self.bboxes = []

        # these are initialized when adding the first volume, see below
        self.affine: np.ndarray = None
        self.shape = None
        self.voxels: np.ndarray = None

    def add_img(self, imgdata: np.ndarray, affine: np.ndarray):

        if self.num_volumes == 0:
            self.affine = affine
            self.shape = imgdata.shape
            self.voxels = np.zeros(imgdata.shape, dtype=np.int32) - 1
        else:
            if (imgdata.shape != self.shape) or ((affine - self.affine).sum() != 0):
                raise RuntimeError(
                    "Building sparse maps from volumes with different voxel spaces is not yet supported in siibra."
                )

        volume = self.num_volumes
        X, Y, Z = [v.astype("int32") for v in np.where(imgdata > 0)]
        for x, y, z, prob in zip(X, Y, Z, imgdata[X, Y, Z]):
            coord_id = self.voxels[x, y, z]
            if coord_id >= 0:
                # Coordinate already seen. Add observed value.
                assert volume not in self.probs[coord_id]
                assert len(self.probs) > coord_id
                self.probs[coord_id][volume] = prob
            else:
                # New coordinate. Append entry with observed value.
                coord_id = len(self.probs)
                self.voxels[x, y, z] = coord_id
                self.probs.append({volume: prob})

        self.bboxes.append(
            {
                "minpoint": (X.min(), Y.min(), Z.min()),
                "maxpoint": (X.max(), Y.max(), Z.max()),
            }
        )

    @property
    def num_volumes(self):
        return len(self.bboxes)

    def max(self):
        return self.voxels.max()

    def coords(self, volume: int):
        # Nx3 array with x/y/z coordinates of the N nonzero values of the given mapindex
        assert volume in range(self.num_volumes)
        coord_ids = [i for i, l in enumerate(self.probs) if volume in l]
        x0, y0, z0 = self.bboxes[volume]["minpoint"]
        x1, y1, z1 = self.bboxes[volume]["maxpoint"]
        return (
            np.array(
                np.where(
                    np.isin(
                        self.voxels[x0: x1 + 1, y0: y1 + 1, z0: z1 + 1],
                        coord_ids,
                    )
                )
            ).T
            + (x0, y0, z0)
        ).T

    def mapped_voxels(self, volume: int):
        # returns the x, y, and z coordinates of nonzero voxels for the map
        # with the given index, together with their corresponding values v.
        assert volume in range(self.num_volumes)
        x, y, z = [v.squeeze() for v in np.split(self.coords(volume), 3)]
        v = [self.probs[i][volume] for i in self.voxels[x, y, z]]
        return x, y, z, v

    @classmethod
    def load(cls, filepath_or_url: str) -> "SparseIndex":
        """
        Loads a precomputed SparseIndex to the memory.

        Parameters
        ----------
        filepath_or_url: str
            Path/url to the SparseIndex files
            (eg. https://url_to_files/basefilename):
            - basefilename.sparseindex_v1.probs.txt.gz
            - basefilename.sparseindex_v1.bboxes.txt.gz
            - basefilename.sparseindex_v1.voxels.nii.gz

        Returns
        -------
        SparseIndex
        """
        from gzip import decompress
        spindtxt_decoder = lambda b: decompress(b).decode('utf-8').strip().splitlines()

        probsfile = filepath_or_url + SparseIndex._SUFFIXES["probs"]
        bboxfile = filepath_or_url + SparseIndex._SUFFIXES["bboxes"]
        voxelfile = filepath_or_url + SparseIndex._SUFFIXES["voxels"]
        if all(path.isfile(f) for f in [probsfile, bboxfile, voxelfile]):
            request = FileLoader
        else:
            request = HttpRequest

        result = cls()

        voxels = request(voxelfile).get()
        result.voxels = np.asanyarray(voxels.dataobj)
        result.affine = voxels.affine
        result.shape = voxels.shape

        lines_probs = request(probsfile, func=spindtxt_decoder).get()
        for line in siibra_tqdm(
            lines_probs,
            total=len(lines_probs),
            desc="Loading sparse index",
            unit="voxels"
        ):
            fields = line.strip().split(" ")
            mapindices = list(map(int, fields[0::2]))
            values = list(map(float, fields[1::2]))
            D = dict(zip(mapindices, values))
            result.probs.append(D)

        lines_bboxes = request(bboxfile, func=spindtxt_decoder).get()
        for line in lines_bboxes:
            fields = line.strip().split(" ")
            result.bboxes.append({
                "minpoint": tuple(map(int, fields[:3])),
                "maxpoint": tuple(map(int, fields[3:])),
            })

        return result

    def save(self, base_filename: str, folder: str = ""):
        """
        Save SparseIndex (3x) files to under the folder `folder`
        with base_filename. If SparseIndex is not cached, siibra will first
        create it first.

        Parameters
        ----------
        base_filename: str
            The files that will be created as:
            - base_filename.sparseindex_v1.probs.txt.gz
            - base_filename.sparseindex_v1.bboxes.txt.gz
            - base_filename.sparseindex_v1.voxels.nii.gz

        folder: str, default=""
        """
        from nibabel import Nifti1Image
        import gzip

        fullpath = path.join(folder, base_filename)
        logger.info(f"Saving SparseIndex to '{base_filename}' with suffixes {SparseIndex._SUFFIXES}")

        if folder and not path.isdir(folder):
            makedirs(folder)

        Nifti1Image(self.voxels, self.affine).to_filename(
            fullpath + SparseIndex._SUFFIXES["voxels"]
        )
        with gzip.open(fullpath + SparseIndex._SUFFIXES["probs"], 'wt') as f:
            for D in self.probs:
                f.write(
                    "{}\n".format(
                        " ".join(f"{i} {p}" for i, p in D.items())
                    )
                )
        with gzip.open(fullpath + SparseIndex._SUFFIXES["bboxes"], "wt") as f:
            for bbox in self.bboxes:
                f.write(
                    "{} {}\n".format(
                        " ".join(map(str, bbox["minpoint"])),
                        " ".join(map(str, bbox["maxpoint"])),
                    )
                )
        logger.info(f"SparseIndex is saved to {fullpath}.")

    @classmethod
    def from_sparsemap(cls, sparsemap: "SparseMap") -> "SparseIndex":
        with provider.SubvolumeProvider.UseCaching():
            spind = cls()
            for img in siibra_tqdm(
                sparsemap.fetch_iter(), total=len(sparsemap), unit="maps",
                desc="Fetching volumetric maps and computing SparseIndex"
            ):
                spind.add_img(np.asanyarray(img.dataobj), img.affine)
        return spind


class SparseMap(parcellationmap.Map):
    """
    A sparse representation of list of statistical (e.g. probabilistic) brain
    region maps.

    It represents the 3D statistical maps of N brain regions by two data structures:

    1) 'spatial_index', a 3D volume where non-negative values represent unique indices into a list of region assignments
    2) 'probs', a list of region assignments where each entry is a dict

    More precisely, given ``i = sparse_index.voxels[x, y, z]`` we define that

    - if `i<0`, no brain region is assigned at this location
    - if `i>=0`, ``probs[i]`` defines the probabilities of brain regions.

    Each entry in probs is a dictionary that represents the region assignments for
    the unique voxel where ``spatial_index == i``. The assignment maps from a MapIndex
    to the actual (probability) value.
    """

    def __init__(
        self,
        identifier: str,
        name: str,
        space_spec: dict,
        parcellation_spec: dict,
        indices: Dict[str, MapIndex],
        volumes: list = [],
        shortname: str = "",
        description: str = "",
        modality: str = None,
        publications: list = [],
        datasets: list = [],
        prerelease: bool = False,
    ):
        parcellationmap.Map.__init__(
            self,
            identifier=identifier,
            name=name,
            space_spec=space_spec,
            parcellation_spec=parcellation_spec,
            indices=indices,
            shortname=shortname,
            description=description,
            modality=modality,
            publications=publications,
            datasets=datasets,
            volumes=volumes,
            prerelease=prerelease,
        )
        self._sparse_index_cached = None

    @property
    def _cache_prefix(self):
        return CACHE.build_filename(f"{self.parcellation.id}_{self.space.id}_{self.maptype}_{self.key.lower()}_index")

    @property
    def sparse_index(self):
        if self._sparse_index_cached is None:
            try:  # try loading from cache on disk
                spind = SparseIndex.load(self._cache_prefix)
            except Exception:
                spind = None
            if spind is None:  # try from precomputed source
                try:
                    logger.info("Downloading and loading precomputed SparseIndex...")
                    spind = SparseIndex.load(SparseIndex._DATAPROXY_BASEURL + self.key.lower())
                except Exception:
                    logger.error("Failed to download precomputed SparseIndex.", exc_info=1)
            if spind is None:  # Download each map and compute the SparseIndex
                spind = SparseIndex.from_sparsemap(self)
                spind.save(self._cache_prefix, folder=CACHE.folder)
            self._sparse_index_cached = spind
        assert self._sparse_index_cached.max() == len(self._sparse_index_cached.probs) - 1
        return self._sparse_index_cached

    @property
    def affine(self):
        return self.sparse_index.affine

    @property
    def shape(self):
        return self.sparse_index.shape

    def _read_voxel(self, x, y, z):
        spind = self.sparse_index
        vx = spind.voxels[x, y, z]
        if isinstance(vx, int):
            return list(
                (None, volume, None, value)
                for volume, value in spind.probs[vx].items()
            )
        else:
            return list(
                (pointindex, volume, None, value)
                for pointindex, voxel in enumerate(vx)
                for volume, value in spind.probs[voxel].items()
            )

    def _assign_volume(
        self,
        queryvolume: "_volume.Volume",
        minsize_voxel: int,
        lower_threshold: float,
        split_components: bool = True
    ) -> List[parcellationmap.AssignImageResult]:
        """
        Assign an image volume to this sparse map.

        Parameters
        -----------
        queryvolume: Volume
            the volume to be compared with maps
        minsize_voxel: int, default: 1
            Minimum voxel size of image components to be taken into account.
        lower_threshold: float, default: 0
            Lower threshold on values in the statistical map. Values smaller than
            this threshold will be excluded from the assignment computation.
        split_components: bool, default: True
            Whether to split the query volume into disjoint components.
        """
        from nibabel import Nifti1Image

        queryimg = queryvolume.fetch()
        assert isinstance(queryimg, Nifti1Image)
        assignments = []

        # resample query image into this image's voxel space, if required (nilearn checks)
        queryimg = resample_img_to_img(
            source_img=queryimg,
            target_img=Nifti1Image(
                np.zeros(self.shape), affine=self.affine, dtype=queryimg.dataobj.dtype
            )
        )
        self.space.get_template()
        querydata = np.asanyarray(queryimg.dataobj).squeeze()

        iter_func = connected_components if split_components else lambda img: [(1, img)]

        for mode, modemask in iter_func(querydata):

            # determine bounding box of the mode
            XYZ2 = np.array(np.where(modemask)).T
            position = np.dot(self.affine, np.r_[XYZ2.mean(0), 1])[:3]
            if XYZ2.shape[0] <= minsize_voxel:
                continue
            X2, Y2, Z2 = [v.squeeze() for v in np.split(XYZ2, 3, axis=1)]

            bbox2 = boundingbox.BoundingBox(XYZ2.min(0), XYZ2.max(0) + 1, space=None)
            if bbox2.volume == 0:
                continue

            spind = self.sparse_index

            for volume in siibra_tqdm(
                range(len(self)),
                desc=f"Assigning structure #{mode} to {len(self)} sparse maps",
                total=len(self),
                unit=" map"
            ):
                bbox1 = boundingbox.BoundingBox(
                    self.sparse_index.bboxes[volume]["minpoint"],
                    self.sparse_index.bboxes[volume]["maxpoint"],
                    space=None,
                )
                if bbox1.intersection(bbox2) is None:
                    continue

                # compute union of voxel space bounding boxes
                bbox = bbox1.union(bbox2)
                bbshape = np.array(bbox.shape, dtype="int") + 1
                x0, y0, z0 = map(int, bbox.minpoint)

                # build flattened vector of map values
                v1 = np.zeros(np.prod(bbshape))
                XYZ1 = spind.coords(volume).T
                X1, Y1, Z1 = [v.squeeze() for v in np.split(XYZ1, 3, axis=1)]
                indices1 = np.ravel_multi_index(
                    (X1 - x0, Y1 - y0, Z1 - z0), bbshape
                )
                v1[indices1] = [spind.probs[i][volume] for i in spind.voxels[X1, Y1, Z1]]
                v1[v1 < lower_threshold] = 0

                # build flattened vector of input image mode
                v2 = np.zeros(np.prod(bbshape))
                indices2 = np.ravel_multi_index(
                    (X2 - x0, Y2 - y0, Z2 - z0), bbshape
                )
                v2[indices2] = querydata[X2, Y2, Z2]

                assert v1.shape == v2.shape

                intersection = np.sum(
                    (v1 > 0) & (v2 > 0)
                )  # np.minimum(v1, v2).sum()
                if intersection == 0:
                    continue
                iou = intersection / np.sum(
                    (v1 > 0) | (v2 > 0)
                )  # np.maximum(v1, v2).sum()

                v1d = v1 - v1.mean()
                v2d = v2 - v2.mean()
                rho = (
                    (v1d * v2d).sum()
                    / np.sqrt((v1d ** 2).sum())
                    / np.sqrt((v2d ** 2).sum())
                )

                maxval = v1.max()

                assignments.append(
                    parcellationmap.AssignImageResult(
                        input_structure=mode,
                        centroid=tuple(position.round(2)),
                        volume=volume,
                        fragment=None,
                        map_value=maxval,
                        intersection_over_union=iou,
                        intersection_over_first=intersection / (v1 > 0).sum(),
                        intersection_over_second=intersection / (v2 > 0).sum(),
                        correlation=rho,
                        weighted_mean_of_first=np.sum(v1 * v2) / np.sum(v2),
                        weighted_mean_of_second=np.sum(v1 * v2) / np.sum(v1)
                    )
                )

        return assignments



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base type of features in tabular formats."""

from zipfile import ZipFile
from typing import Callable

import pandas as pd
from textwrap import wrap

from .. import feature
from .. import anchor as _anchor
from ...commons import logger
from ...retrieval import requests


class Tabular(feature.Feature, category="generic", configuration_folder="features/tabular"):
    """
    Represents a table of different measures anchored to a brain location.

    Columns represent different types of values, while rows represent different
    samples. The number of columns might thus be interpreted as the feature
    dimension.

    As an example, receptor fingerprints use rows to represent different
    neurotransmitter receptors, and separate columns for the mean and standard
    deviations measure across multiple tissue samples.
    """

    def __init__(
        self,
        description: str,
        modality: str,
        anchor: _anchor.AnatomicalAnchor,
        file: str = None,
        decoder: Callable = None,
        data: pd.DataFrame = None,  # sample x feature dimension
        datasets: list = [],
        id: str = None,
        prerelease: bool = False,
    ):
        feature.Feature.__init__(
            self,
            modality=modality,
            description=description,
            anchor=anchor,
            datasets=datasets,
            id=id,
            prerelease=prerelease
        )
        self._loader = None if file is None else requests.HttpRequest(file, func=decoder)
        if file is not None:
            assert data is None
        self._data_cached = data

    @property
    def data(self):
        if self._loader is not None:
            self._data_cached = self._loader.get()
        return self._data_cached.copy()

    def _to_zip(self, fh: ZipFile):
        super()._to_zip(fh)
        fh.writestr("tabular.csv", self.data.to_csv())

    def plot(self, *args, backend="matplotlib", **kwargs):
        """
        Create a bar plot of a columns of the data.
        Parameters
        ----------
        backend: str
            "matplotlib", "plotly", or others supported by pandas DataFrame
            plotting backend.
        **kwargs
            takes Matplotlib.pyplot keyword arguments
        """
        wrapwidth = kwargs.pop("textwrap") if "textwrap" in kwargs else 40
        kwargs["title"] = kwargs.get(
            "title",
            "\n".join(wrap(
                f"{self.modality} in {', '.join({_.name for _ in self.anchor.regions})}",
                wrapwidth
            ))
        )
        kwargs["kind"] = kwargs.get("kind", "bar")
        kwargs["y"] = kwargs.get("y", self.data.columns[0])
        if backend == "matplotlib":
            try:
                import matplotlib.pyplot as plt
            except ImportError as e:
                logger.error(
                    "matplotlib not available. Please install matplotlib or use or another backend such as plotly."
                )
                raise e
            # default kwargs
            if kwargs.get("error_y") is None:
                kwargs["yerr"] = kwargs.get("yerr", 'std' if 'std' in self.data.columns else None)
                yerr_label = f" \u00b1 {kwargs.get('yerr')}" if kwargs.get('yerr') else ''
            if kwargs.get('kind') == 'bar':
                kwargs["width"] = kwargs.get("width", 0.8)
                kwargs["edgecolor"] = kwargs.get('edgecolor', 'black')
                kwargs["linewidth"] = kwargs.get('linewidth', 1.0)
                kwargs["capsize"] = kwargs.get('capsize', 4)
            kwargs["ylabel"] = kwargs.get(
                "ylabel",
                f"{kwargs['y']}{yerr_label}" + f"\n{self.unit}" if hasattr(self, 'unit') else ""
            )
            kwargs["grid"] = kwargs.get("grid", True)
            kwargs["legend"] = kwargs.get("legend", False)
            kwargs["color"] = kwargs.get('color', 'darkgrey')

            xticklabel_rotation = kwargs.get("rot", 60)
            ax = self.data.plot(*args, backend=backend, **kwargs)
            ax.set_title(ax.get_title(), fontsize="medium")
            ax.set_xticklabels(
                ax.get_xticklabels(),
                rotation=xticklabel_rotation,
                ha='center' if xticklabel_rotation % 90 == 0 else 'right'
            )
            ax.spines['top'].set_visible(False)
            ax.spines['right'].set_visible(False)
            plt.tight_layout()
            return ax
        elif backend == "plotly":
            kwargs["title"] = kwargs["title"].replace("\n", "<br>")
            kwargs["error_y"] = kwargs.get("error_y", 'std' if 'std' in self.data.columns else None)
            error_y_label = f" &plusmn; {kwargs.get('error_y')}<br>" if kwargs.get('error_y') else ''
            kwargs["labels"] = {
                "index": kwargs.pop("xlabel", None) or kwargs.pop("index", ""),
                "value": kwargs.pop("ylabel", None) or kwargs.pop(
                    "value",
                    f"{kwargs.get('y')}{error_y_label} {self.unit if hasattr(self, 'unit') else ''}"
                )
            }
            fig = self.data.plot(*args, backend=backend, **kwargs)
            fig.update_layout(
                yaxis_title=kwargs["labels"]['value'],
                title=dict(
                    automargin=True, yref="container", xref="container",
                    pad=dict(t=40), xanchor="left", yanchor="top"
                )
            )
            return fig
        else:
            return self.data.plot(*args, backend=backend, **kwargs)



# Copyright 2018-2025
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Union, Dict, Tuple

import numpy as np
import nibabel as nib

from . import provider as _provider
from ...commons import logger, resample_img_to_img
from ...retrieval import requests
from ...locations import pointcloud, boundingbox as _boundingbox


class NiftiProvider(_provider.VolumeProvider, srctype="nii"):

    def __init__(self, src: Union[str, Dict[str, str], nib.Nifti1Image, Tuple[np.ndarray, np.ndarray]]):
        """
        Construct a new NIfTI volume source, from url, local file, or Nift1Image object.
        """
        _provider.VolumeProvider.__init__(self)

        self._init_url: Union[str, Dict[str, str]] = None

        def loader(url):
            if os.path.isfile(url):
                return lambda fn=url: nib.load(fn)
            else:
                req = requests.HttpRequest(url)
                return lambda req=req: req.data

        if isinstance(src, nib.Nifti1Image):
            self._img_loaders = {None: lambda img=src: img}
        elif isinstance(src, str):  # one single image to load
            self._img_loaders = {None: loader(src)}
        elif isinstance(src, dict):  # assuming multiple for fragment images
            self._img_loaders = {lbl: loader(url) for lbl, url in src.items()}
        elif isinstance(src, tuple):
            assert len(src) == 2
            assert all(isinstance(_, np.ndarray) for _ in src)
            self._img_loaders = {None: lambda data=src[0], affine=src[1]: nib.Nifti1Image(data, affine)}
        else:
            raise ValueError(f"Invalid source specification for {self.__class__}: {src}")
        if not isinstance(src, (nib.Nifti1Image, tuple)):
            self._init_url = src

    @property
    def _url(self) -> Union[str, Dict[str, str]]:
        return self._init_url

    @property
    def fragments(self):
        return [k for k in self._img_loaders if k is not None]

    def get_boundingbox(self, **fetch_kwargs) -> "_boundingbox.BoundingBox":
        """
        Return the bounding box in physical coordinates of the union of
        fragments in this nifti volume.

        Parameters
        ----------
        fetch_kwargs:
            Not used
        """
        bbox = None
        for loader in self._img_loaders.values():
            img = loader()
            if len(img.shape) > 3:
                logger.warning(
                    f"N-D NIfTI volume has shape {img.shape}, but "
                    f"bounding box considers only {img.shape[:3]}"
                )
            shape = img.shape[:3]
            next_bbox = _boundingbox.BoundingBox(
                (0, 0, 0), shape, space=None
            ).transform(img.affine)
            bbox = next_bbox if bbox is None else bbox.union(next_bbox)
        return bbox

    def _merge_fragments(self) -> nib.Nifti1Image:
        """
        Merge all fragments this volume contains into one Nifti1Image.
        """
        bbox = self.get_boundingbox(clip=False, background=0.0)
        num_conflicts = 0
        result = None
        for loader in self._img_loaders.values():
            img = loader()
            if result is None:
                # build the empty result image with its own affine and voxel space
                s0 = np.identity(4)
                s0[:3, -1] = list(bbox.minpoint.transform(np.linalg.inv(img.affine)))
                result_affine = np.dot(img.affine, s0)  # adjust global bounding box offset to get global affine
                voxdims = np.asanyarray(np.ceil(
                    bbox.transform(np.linalg.inv(result_affine)).shape
                ), dtype="int")
                result_arr = np.zeros(voxdims, dtype=img.dataobj.dtype)
                result = nib.Nifti1Image(dataobj=result_arr, affine=result_affine)

            # resample to merge template and update it
            resampled_img = resample_img_to_img(source_img=img, target_img=result)
            arr = np.asanyarray(resampled_img.dataobj)
            nonzero_voxels = arr != 0
            num_conflicts += np.count_nonzero(result_arr[nonzero_voxels])
            result_arr[nonzero_voxels] = arr[nonzero_voxels]

        if num_conflicts > 0:
            num_voxels = np.count_nonzero(result_arr)
            logger.warning(
                f"Merging fragments required to overwrite {num_conflicts} "
                f"conflicting voxels ({num_conflicts / num_voxels * 100.:2.1f}%)."
            )

        return result

    def fetch(
        self,
        fragment: str = None,
        voi: _boundingbox.BoundingBox = None,
        label: int = None
    ):
        """
        Loads and returns a Nifti1Image object

        Parameters
        ----------
        fragment: str
            Optional name of a fragment volume to fetch, if any.
            For example, some volumes are split into left and right hemisphere fragments.
            see :func:`~siibra.volumes.Volume.fragments`
        voi : BoundingBox
            optional specification of a volume of interest to fetch.
        label: int, default: None
            Optional: a label index can be provided. Then the mask of the
            3D volume will be returned, where voxels matching this label
            are marked as "1".
        """

        result = None
        if len(self._img_loaders) > 1:
            if fragment is None:
                logger.info(
                    f"Merging fragments [{', '.join(self._img_loaders.keys())}]. "
                    f"You can select one using {self.__class__.__name__}.fetch(fragment=<name>)."
                )
                result = self._merge_fragments()
            else:
                matched_names = [n for n in self._img_loaders if fragment.lower() in n.lower()]
                if len(matched_names) != 1:
                    raise ValueError(
                        f"Requested fragment '{fragment}' could not be matched uniquely "
                        f"to [{', '.join(self._img_loaders)}]"
                    )
                else:
                    result = self._img_loaders[matched_names[0]]()
        else:
            assert len(self._img_loaders) > 0
            fragment_name, loader = next(iter(self._img_loaders.items()))
            if (fragment_name is not None) and (fragment is not None):
                assert fragment.lower() in fragment_name.lower()
            result = loader()

        if voi is not None:
            zoom_xyz = np.array(result.header.get_zooms())  # voxel dimensions in xyzt_units
            bb_vox = voi.transform(np.linalg.inv(result.affine))
            x0, y0, z0 = np.floor(np.array(bb_vox.minpoint.coordinate) / zoom_xyz).astype(int)
            x1, y1, z1 = np.ceil(np.array(bb_vox.maxpoint.coordinate) / zoom_xyz).astype(int)
            shift = np.identity(4)
            shift[:3, -1] = bb_vox.minpoint
            result = nib.Nifti1Image(
                dataobj=result.dataobj[x0:x1, y0:y1, z0:z1],
                affine=np.dot(result.affine, shift),
                dtype=result.header.get_data_dtype(),
            )

        if label is not None:
            result = nib.Nifti1Image(
                (result.get_fdata() == label).astype('uint8'),
                result.affine,
                dtype='uint8'
            )

        return result

    def get_shape(self, resolution_mm=None):
        if resolution_mm is not None:
            raise NotImplementedError(
                "NiftiVolume does not support to specify different image resolutions"
            )
        try:
            loader_shapes = {loader().shape for loader in self._img_loaders.values()}
            if len(loader_shapes) == 1:
                return next(iter(loader_shapes))
            else:
                raise RuntimeError(f"Fragments have different shapes: {loader_shapes}")
        except AttributeError as e:
            logger.error(
                f"Invalid object type/s {[type(loader()) for loader in self._img_loaders.values()]} of image for {self}."
            )
            raise (e)

    def is_float(self):
        return all(
            loader().dataobj.dtype.kind == "f"
            for loader in self._img_loaders.values()
        )

    def find_peaks(self, min_distance_mm=5):
        """
        Find peaks in the image data.

        Parameters
        ----------
        min_distance_mm : float
            Minimum distance between peaks in mm

        Returns:
        --------
        PointCloud
        """

        from skimage.feature.peak import peak_local_max
        from ...commons import affine_scaling

        img = self.fetch()
        dist = int(min_distance_mm / affine_scaling(img.affine) + 0.5)
        voxels = peak_local_max(
            img.get_fdata(),
            exclude_border=False,
            min_distance=dist,
        )
        return (
            pointcloud.PointCloud(
                [np.dot(img.affine, [x, y, z, 1])[:3] for x, y, z in voxels],
                space=self.space,
            ),
            img,
        )


class ZipContainedNiftiProvider(NiftiProvider, srctype="zip/nii"):

    def __init__(self, src: str):
        """
        Construct a new NIfTI volume source, from url, local file, or Nift1Image object.
        """
        _provider.VolumeProvider.__init__(self)
        zipurl, zipped_file = src.split(" ")
        req = requests.ZipfileRequest(zipurl, zipped_file)
        self._img_loaders = {None: lambda req=req: req.data}

        # required for self._url property
        self._init_url = src
